""" Frontend to FTS3 MySQL DB. Written using sqlalchemy
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# We disable the no-member error because
# they are constructed by SQLAlchemy for all
# the objects mapped to a table.
# pylint: disable=no-member

import datetime
import errno

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.sql.expression import and_
from sqlalchemy.orm import relationship, sessionmaker, mapper
from sqlalchemy.sql import update, delete
from sqlalchemy import (
    create_engine,
    Table,
    Column,
    MetaData,
    ForeignKey,
    Integer,
    String,
    DateTime,
    Enum,
    BigInteger,
    SmallInteger,
    Float,
    func,
    text,
)

# # from DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.DataManagementSystem.Client.FTS3Operation import FTS3Operation, FTS3TransferOperation, FTS3StagingOperation
from DIRAC.DataManagementSystem.Client.FTS3File import FTS3File
from DIRAC.DataManagementSystem.Client.FTS3Job import FTS3Job
from DIRAC.ConfigurationSystem.Client.Utilities import getDBParameters

__RCSID__ = "$Id$"


metadata = MetaData()


fts3FileTable = Table(
    "Files",
    metadata,
    Column("fileID", Integer, primary_key=True),
    Column("operationID", Integer, ForeignKey("Operations.operationID", ondelete="CASCADE"), nullable=False),
    Column("attempt", Integer, server_default="0"),
    Column("lastUpdate", DateTime, onupdate=func.utc_timestamp()),
    Column("rmsFileID", Integer, server_default="0"),
    Column("lfn", String(1024)),
    Column("checksum", String(255)),
    Column("size", BigInteger),
    Column("targetSE", String(255), nullable=False),
    Column("error", String(2048)),
    Column("ftsGUID", String(255)),
    Column("status", Enum(*FTS3File.ALL_STATES), server_default=FTS3File.INIT_STATE, index=True),
    mysql_engine="InnoDB",
)

mapper(FTS3File, fts3FileTable)


fts3JobTable = Table(
    "Jobs",
    metadata,
    Column("jobID", Integer, primary_key=True),
    Column("operationID", Integer, ForeignKey("Operations.operationID", ondelete="CASCADE"), nullable=False),
    Column("submitTime", DateTime),
    Column("lastUpdate", DateTime, onupdate=func.utc_timestamp()),
    Column("lastMonitor", DateTime),
    Column("completeness", Float),
    # Could be fetched from Operation, but bad for perf
    Column("username", String(255)),
    # Could be fetched from Operation, but bad for perf
    Column("userGroup", String(255)),
    Column("ftsGUID", String(255)),
    Column("ftsServer", String(255)),
    Column("error", String(2048)),
    Column("status", Enum(*FTS3Job.ALL_STATES), server_default=FTS3Job.INIT_STATE, index=True),
    Column("assignment", String(255), server_default=None),
    mysql_engine="InnoDB",
)

mapper(FTS3Job, fts3JobTable)


fts3OperationTable = Table(
    "Operations",
    metadata,
    Column("operationID", Integer, primary_key=True),
    Column("username", String(255)),
    Column("userGroup", String(255)),
    # -1 because with 0 we would get any random request
    # when performing reqClient.getRequest
    Column("rmsReqID", Integer, server_default="-1"),
    Column("rmsOpID", Integer, server_default="0", index=True),
    Column("sourceSEs", String(255)),
    Column("activity", String(255)),
    Column("priority", SmallInteger),
    Column("creationTime", DateTime),
    Column("lastUpdate", DateTime, onupdate=func.utc_timestamp()),
    Column("status", Enum(*FTS3Operation.ALL_STATES), server_default=FTS3Operation.INIT_STATE, index=True),
    Column("error", String(1024)),
    Column("type", String(255)),
    Column("assignment", String(255), server_default=None),
    mysql_engine="InnoDB",
)


fts3Operation_mapper = mapper(
    FTS3Operation,
    fts3OperationTable,
    properties={
        "ftsFiles": relationship(
            FTS3File,
            lazy="joined",  # Immediately load the entirety of the object
            innerjoin=True,  # Use inner join instead of left outer join
            cascade="all, delete-orphan",  # if a File is removed from the list,
            # remove it from the DB
            passive_deletes=True,  # used together with cascade='all, delete-orphan'
        ),
        "ftsJobs": relationship(
            FTS3Job,
            lazy="subquery",  # Immediately load the entirety of the object,
            # but use a subquery to do it
            # This is to avoid the cartesian product between the three tables.
            # https://docs.sqlalchemy.org/en/latest/orm/loading_relationships.html#subquery-eager-loading
            cascade="all, delete-orphan",  # if a File is removed from the list,
            # remove it from the DB
            passive_deletes=True,  # used together with cascade='all, delete-orphan'
        ),
    },
    polymorphic_on="type",
    polymorphic_identity="Abs",
)

mapper(FTS3TransferOperation, fts3OperationTable, inherits=fts3Operation_mapper, polymorphic_identity="Transfer")

mapper(FTS3StagingOperation, fts3OperationTable, inherits=fts3Operation_mapper, polymorphic_identity="Staging")


# About synchronize_session:
# The FTS3DB class uses SQLAlchemy in a mixed mode:
# both the ORM and the Core style.
# Up to 1.3, the `session.execute` statements had no ORM functionality,
# meaning that the session cache were not updated when using `update` or `delete`
# Although it could be an issue, it was not really one in our case, since we always close
# the session
# As of 1.4, `session.execute` supports ORM functionality, and thus needs more info to know
# how to manage `update` or `delete`. Hense the `synchronize_session` option.
# We set it to `False` simply because we do not rely on the session cache.
# Please see https://github.com/sqlalchemy/sqlalchemy/discussions/6159 for detailed discussion

########################################################################
class FTS3DB(object):
    """
    .. class:: RequestDB

    db holding requests
    """

    def __getDBConnectionInfo(self, fullname):
        """Collect from the CS all the info needed to connect to the DB.
        This should be in a base class eventually
        """

        result = getDBParameters(fullname)
        if not result["OK"]:
            raise Exception("Cannot get database parameters: %s" % result["Message"])

        dbParameters = result["Value"]
        self.dbHost = dbParameters["Host"]
        self.dbPort = dbParameters["Port"]
        self.dbUser = dbParameters["User"]
        self.dbPass = dbParameters["Password"]
        self.dbName = dbParameters["DBName"]

    def __init__(self, pool_size=15):
        """c'tor

        :param self: self reference
        :param pool_size: size of the connection pool to the DB

        """

        self.log = gLogger.getSubLogger("FTS3DB")
        # Initialize the connection info
        self.__getDBConnectionInfo("DataManagement/FTS3DB")

        runDebug = gLogger.getLevel() == "DEBUG"
        self.engine = create_engine(
            "mysql://%s:%s@%s:%s/%s" % (self.dbUser, self.dbPass, self.dbHost, self.dbPort, self.dbName),
            echo=runDebug,
            pool_size=pool_size,
            pool_recycle=3600,
        )

        metadata.bind = self.engine

        self.dbSession = sessionmaker(bind=self.engine)

    def createTables(self):
        """create tables"""
        try:
            metadata.create_all(self.engine)
        except SQLAlchemyError as e:
            return S_ERROR(e)
        return S_OK()

    def persistOperation(self, operation):
        """update or insert request into db
            Also release the assignment tag

        :param operation: FTS3Operation instance
        """

        session = self.dbSession(expire_on_commit=False)

        # set the assignment to NULL
        # so that another agent can work on the request
        operation.assignment = None
        # because of the merge we have to explicitely set lastUpdate
        operation.lastUpdate = func.utc_timestamp()
        try:

            # Merge it in case it already is in the DB
            operation = session.merge(operation)
            session.add(operation)
            session.commit()
            session.expunge_all()

            return S_OK(operation.operationID)

        except SQLAlchemyError as e:
            session.rollback()
            self.log.exception("persistOperation: unexpected exception", lException=e)
            return S_ERROR("persistOperation: unexpected exception %s" % e)
        finally:
            session.close()

    def getOperation(self, operationID):
        """read request

          This does not set the assignment flag

        :param operationID: ID of the FTS3Operation

        """

        # expire_on_commit is set to False so that we can still use the object
        # after we close the session
        session = self.dbSession(expire_on_commit=False)

        try:

            operation = session.query(FTS3Operation).filter(getattr(FTS3Operation, "operationID") == operationID).one()

            session.commit()

            ###################################
            session.expunge_all()
            return S_OK(operation)

        except NoResultFound as e:
            # We use the ENOENT error, even if not really a file error :)
            return S_ERROR(errno.ENOENT, "No FTS3Operation with id %s" % operationID)
        except SQLAlchemyError as e:
            return S_ERROR("getOperation: unexpected exception : %s" % e)
        finally:
            session.close()

    def getActiveJobs(self, limit=20, lastMonitor=None, jobAssignmentTag="Assigned"):
        """Get  the FTSJobs that are not in a final state, and are not assigned for monitoring
         or has its operation being treated

         By assigning the job to the DB:
           * it cannot be monitored by another agent
           * the operation to which it belongs cannot be treated

        :param limit: max number of Jobs to retrieve
        :param lastMonitor: jobs monitored earlier than the given date
        :param jobAssignmentTag: if not None, block the Job for other queries,
                               and use it as a prefix for the value in the operation table

        :returns: list of FTS3Jobs

        """

        session = self.dbSession(expire_on_commit=False)

        try:
            # the tild sign is for "not"

            ftsJobsQuery = (
                session.query(FTS3Job)
                .join(FTS3Operation)
                .filter(~FTS3Job.status.in_(FTS3Job.FINAL_STATES))
                .filter(FTS3Job.assignment.is_(None))
                .filter(FTS3Operation.assignment.is_(None))
            )

            if lastMonitor:
                ftsJobsQuery = ftsJobsQuery.filter(FTS3Job.lastMonitor < lastMonitor)

            if jobAssignmentTag:
                ftsJobsQuery = ftsJobsQuery.with_for_update()

            ftsJobsQuery = ftsJobsQuery.order_by(FTS3Job.lastMonitor.asc())
            ftsJobsQuery = ftsJobsQuery.limit(limit)

            ftsJobs = ftsJobsQuery.all()

            if jobAssignmentTag:
                jobAssignmentTag += "_%s" % datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")

                jobIds = [job.jobID for job in ftsJobs]
                if jobIds:
                    session.execute(
                        update(FTS3Job)
                        .where(FTS3Job.jobID.in_(jobIds))
                        .values({"assignment": jobAssignmentTag})
                        .execution_options(synchronize_session=False)  # see comment about synchronize_session
                    )

            session.commit()

            session.expunge_all()

            return S_OK(ftsJobs)

        except SQLAlchemyError as e:
            session.rollback()
            return S_ERROR("getAllActiveJobs: unexpected exception : %s" % e)
        finally:
            session.close()

    def updateFileStatus(self, fileStatusDict, ftsGUID=None):
        """Update the file ftsStatus and error
         The update is only done if the file is not in a final state
         (To avoid bringing back to life a file by consuming MQ a posteriori)



         TODO: maybe it should query first the status and filter the rows I want to update !

        :param fileStatusDict: { fileID : { status , error, ftsGUID } }
        :param ftsGUID: If specified, only update the rows where the ftsGUID matches this value.
                        This avoids two jobs handling the same file one after another to step on each other foot.
                        Note that for the moment it is an optional parameter, but it may turn mandatory soon.

        """

        # This here is inneficient as we update every files, even if it did not change, and we commit every time.
        # It would probably be best to update only the files that changed.
        # However, commiting every time is the recommendation of MySQL
        # (https://dev.mysql.com/doc/refman/5.7/en/innodb-deadlocks-handling.html)

        for fileID, valueDict in fileStatusDict.items():

            session = self.dbSession()
            try:

                updateDict = {FTS3File.status: valueDict["status"]}

                # We only update error if it is specified
                if "error" in valueDict:
                    newError = valueDict["error"]
                    # Replace empty string with None
                    if not newError:
                        newError = None
                    updateDict[FTS3File.error] = newError

                # We only update ftsGUID if it is specified
                if "ftsGUID" in valueDict:
                    newFtsGUID = valueDict["ftsGUID"]
                    # Replace empty string with None
                    if not newFtsGUID:
                        newFtsGUID = None
                    updateDict[FTS3File.ftsGUID] = newFtsGUID

                # We only update the lines matching:
                # * the good fileID
                # * the status is not Final

                whereConditions = [FTS3File.fileID == fileID, ~FTS3File.status.in_(FTS3File.FINAL_STATES)]

                # If an ftsGUID is specified, add it to the `where` condition
                if ftsGUID:
                    whereConditions.append(FTS3File.ftsGUID == ftsGUID)

                updateQuery = (
                    update(FTS3File)
                    .where(and_(*whereConditions))
                    .values(updateDict)
                    .execution_options(synchronize_session=False)
                )  # see comment about synchronize_session

                session.execute(updateQuery)

                session.commit()

            except SQLAlchemyError as e:
                session.rollback()
                self.log.exception("updateFileFtsStatus: unexpected exception", lException=e)
                return S_ERROR("updateFileFtsStatus: unexpected exception %s" % e)
            finally:
                session.close()

        return S_OK()

    def updateJobStatus(self, jobStatusDict):
        """Update the job Status and error
         The update is only done if the job is not in a final state
         The assignment flag is released

        :param jobStatusDict: { jobID : { status , error, completeness } }
        """
        session = self.dbSession()
        try:

            for jobID, valueDict in jobStatusDict.items():

                updateDict = {FTS3Job.status: valueDict["status"]}

                # We only update error if it is specified
                if "error" in valueDict:
                    newError = valueDict["error"]
                    # Replace empty string with None
                    if not newError:
                        newError = None
                    updateDict[FTS3Job.error] = newError

                if "completeness" in valueDict:
                    updateDict[FTS3Job.completeness] = valueDict["completeness"]

                if valueDict.get("lastMonitor"):
                    updateDict[FTS3Job.lastMonitor] = func.utc_timestamp()

                updateDict[FTS3Job.assignment] = None

                session.execute(
                    update(FTS3Job)
                    .where(and_(FTS3Job.jobID == jobID, ~FTS3Job.status.in_(FTS3Job.FINAL_STATES)))
                    .values(updateDict)
                    .execution_options(synchronize_session=False)  # see comment about synchronize_session
                )
            session.commit()

            return S_OK()

        except SQLAlchemyError as e:
            session.rollback()
            self.log.exception("updateJobStatus: unexpected exception", lException=e)
            return S_ERROR("updateJobStatus: unexpected exception %s" % e)
        finally:
            session.close()

    def cancelNonExistingJob(self, operationID, ftsGUID):
        """
        Cancel an FTS3Job with the associated FTS3Files.
        This is to be used when the job is not found on the server when monitoring.

        The status of the job and files will be 'Canceled'.
        The error is specifying that the job is not found.
        The ftsGUID of the file is released as to be able to pick it up again

        :param operationID: guess
        :param ftsGUID: guess

        :returns: S_OK() if successful, S_ERROR otherwise
        """
        session = self.dbSession()
        try:

            # We update both the rows of the Jobs and the Files tables
            # having matching operationID and ftsGUID
            # https://docs.sqlalchemy.org/en/13/core/tutorial.html#multiple-table-updates

            # We do not need to specify that the File status should be final
            # since they would have been updated before and the ftsGUID already removed
            updStmt = (
                update(FTS3Job)
                .values(
                    {
                        FTS3File.status: "Canceled",
                        FTS3File.ftsGUID: None,
                        FTS3File.error: "Job %s not found" % ftsGUID,
                        FTS3Job.status: "Canceled",
                        FTS3Job.error: "Job %s not found" % ftsGUID,
                    }
                )
                .where(
                    and_(
                        FTS3File.operationID == FTS3Job.operationID,
                        FTS3File.ftsGUID == FTS3Job.ftsGUID,
                        FTS3Job.operationID == operationID,
                        FTS3Job.ftsGUID == ftsGUID,
                    )
                )
                .execution_options(synchronize_session=False)
            )  # see comment about synchronize_session

            session.execute(updStmt)
            session.commit()

            return S_OK()

        except SQLAlchemyError as e:
            session.rollback()
            self.log.exception("cancelNonExistingJob: unexpected exception", lException=e)
            return S_ERROR("cancelNonExistingJob: unexpected exception %s" % e)
        finally:
            session.close()

    def getNonFinishedOperations(self, limit=20, operationAssignmentTag="Assigned"):
        """Get all the non assigned FTS3Operations that are not yet finished, so either Active or Processed.
        An operation won't be picked if it is already assigned, or one of its job is.

        :param limit: max number of operations to retrieve
        :param operationAssignmentTag: if not None, block the operations for other queries,
                              and use it as a prefix for the value in the operation table
        :return: list of Operations
        """

        session = self.dbSession(expire_on_commit=False)

        try:

            ftsOperations = []

            # We need to do the select in two times because the join clause that makes the limit difficult
            operationIDsQuery = (
                session.query(FTS3Operation.operationID)
                .outerjoin(FTS3Job)
                .filter(FTS3Operation.status.in_(["Active", "Processed"]))
                .filter(FTS3Operation.assignment.is_(None))
                .filter(FTS3Job.assignment.is_(None))
                .order_by(FTS3Operation.lastUpdate.asc())
                .limit(limit)
                .distinct()
            )

            # Block the Operations for other requests
            if operationAssignmentTag:
                operationIDsQuery = operationIDsQuery.with_for_update()

            operationIDs = operationIDsQuery.all()

            operationIDs = [oidTuple[0] for oidTuple in operationIDs]

            if operationIDs:
                # Fetch the operation object for these IDs
                ftsOperations = session.query(FTS3Operation).filter(FTS3Operation.operationID.in_(operationIDs)).all()

                if operationAssignmentTag:
                    operationAssignmentTag += "_%s" % datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")

                    session.execute(
                        update(FTS3Operation)
                        .where(FTS3Operation.operationID.in_(operationIDs))
                        .values({"assignment": operationAssignmentTag})
                        .execution_options(synchronize_session=False)  # see comment about synchronize_session
                    )

            session.commit()
            session.expunge_all()

            return S_OK(ftsOperations)

        except SQLAlchemyError as e:
            session.rollback()
            return S_ERROR("getAllProcessedOperations: unexpected exception : %s" % e)
        finally:
            session.close()

    def kickStuckOperations(self, limit=20, kickDelay=2):
        """finds operations that have not been updated for more than a given
          time but are still assigned and resets the assignment

        :param int limit: number of operations to treat
        :param int kickDelay: age of the lastUpdate in hours
        :returns: S_OK/S_ERROR with number of kicked operations

        """

        session = self.dbSession(expire_on_commit=False)

        try:

            ftsOps = (
                session.query(FTS3Operation.operationID)
                .filter(
                    FTS3Operation.lastUpdate
                    < (func.date_sub(func.utc_timestamp(), text("INTERVAL %d HOUR" % kickDelay)))
                )
                .filter(~FTS3Operation.assignment.is_(None))
                .limit(limit)
            )

            opIDs = [opTuple[0] for opTuple in ftsOps]
            rowCount = 0

            if opIDs:
                result = session.execute(
                    update(FTS3Operation)
                    .where(FTS3Operation.operationID.in_(opIDs))
                    .where(
                        FTS3Operation.lastUpdate
                        < (func.date_sub(func.utc_timestamp(), text("INTERVAL %d HOUR" % kickDelay)))
                    )
                    .values({"assignment": None})
                    .execution_options(synchronize_session=False)  # see comment about synchronize_session
                )
                rowCount = result.rowcount

            session.commit()
            session.expunge_all()

            return S_OK(rowCount)

        except SQLAlchemyError as e:
            session.rollback()
            return S_ERROR("kickStuckOperations: unexpected exception : %s" % e)
        finally:
            session.close()

    def kickStuckJobs(self, limit=20, kickDelay=2):
        """finds jobs that have not been updated for more than a given
          time but are still assigned and resets the assignment

        :param int limit: number of jobs to treat
        :param int kickDelay: age of the lastUpdate in hours
        :returns: S_OK/S_ERROR with number of kicked jobs

        """

        session = self.dbSession(expire_on_commit=False)

        try:

            ftsJobs = (
                session.query(FTS3Job.jobID)
                .filter(
                    FTS3Job.lastUpdate < (func.date_sub(func.utc_timestamp(), text("INTERVAL %d HOUR" % kickDelay)))
                )
                .filter(~FTS3Job.assignment.is_(None))
                .limit(limit)
            )

            jobIDs = [jobTuple[0] for jobTuple in ftsJobs]
            rowCount = 0

            if jobIDs:
                result = session.execute(
                    update(FTS3Job)
                    .where(FTS3Job.jobID.in_(jobIDs))
                    .where(
                        FTS3Job.lastUpdate < (func.date_sub(func.utc_timestamp(), text("INTERVAL %d HOUR" % kickDelay)))
                    )
                    .values({"assignment": None})
                    .execution_options(synchronize_session=False)  # see comment about synchronize_session
                )
                rowCount = result.rowcount

            session.commit()
            session.expunge_all()

            return S_OK(rowCount)

        except SQLAlchemyError as e:
            session.rollback()
            return S_ERROR("kickStuckJobs: unexpected exception : %s" % e)
        finally:
            session.close()

    def deleteFinalOperations(self, limit=20, deleteDelay=180):
        """deletes operation in final state that are older than given time

        :param int limit: number of operations to treat
        :param int deleteDelay: age of the lastUpdate in days
        :returns: S_OK/S_ERROR with number of deleted operations
        """

        session = self.dbSession(expire_on_commit=False)

        try:

            ftsOps = (
                session.query(FTS3Operation.operationID)
                .filter(
                    FTS3Operation.lastUpdate
                    < (func.date_sub(func.utc_timestamp(), text("INTERVAL %d DAY" % deleteDelay)))
                )
                .filter(FTS3Operation.status.in_(FTS3Operation.FINAL_STATES))
                .limit(limit)
            )

            opIDs = [opTuple[0] for opTuple in ftsOps]
            rowCount = 0
            if opIDs:
                result = session.execute(
                    delete(FTS3Operation)
                    .where(FTS3Operation.operationID.in_(opIDs))
                    .execution_options(synchronize_session=False)
                )
                rowCount = result.rowcount

            session.commit()
            session.expunge_all()

            return S_OK(rowCount)

        except SQLAlchemyError as e:
            session.rollback()
            return S_ERROR("deleteFinalOperations: unexpected exception : %s" % e)
        finally:
            session.close()

    def getOperationsFromRMSOpID(self, rmsOpID):
        """Returns the FTS3Operations matching a given RMS OperationID

          This does not set the assignment flag

        :param rmsOpID: ID of the RMS Operation

        """

        # expire_on_commit is set to False so that we can still use the object
        # after we close the session
        session = self.dbSession(expire_on_commit=False)

        try:

            operations = session.query(FTS3Operation).filter(FTS3Operation.rmsOpID == rmsOpID).all()

            session.commit()

            ###################################
            session.expunge_all()
            return S_OK(operations)

        except NoResultFound as e:
            # If there is no such operation, return an empty list
            return S_OK([])
        except SQLAlchemyError as e:
            return S_ERROR("getOperationsFromRMSOpID: unexpected exception : %s" % e)
        finally:
            session.close()



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-lfn-metadata
# Author :  Stuart Paterson
########################################################################
"""
Obtain replica metadata from file catalogue client.

Example:
  $ dirac-dms-lfn-metadata /formation/user/v/vhamar/test.txt
  {'Failed': {},
   'Successful': {'/formation/user/v/vhamar/test.txt': {'Checksum': 'eed20d47',
                                                        'ChecksumType': 'Adler32',
                                                        'CreationDate': datetime.datetime(2011, 2, 11, 14, 52, 47),
                                                        'FileID': 250L,
                                                        'GID': 2,
                                                        'GUID': 'EDE6DDA4-3344-3F39-A993-8349BA41EB23',
                                                        'Mode': 509,
                                                        'ModificationDate': datetime.datetime(2011, 2, 11, 14, 52, 47),
                                                        'Owner': 'vhamar',
                                                        'OwnerGroup': 'dirac_user',
                                                        'Size': 34L,
                                                        'Status': 1,
                                                        'UID': 2}}}
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["LFN:      Logical File Name or file containing LFNs"])
    _, lfns = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.Dirac import Dirac

    dirac = Dirac()
    exitCode = 0
    errorList = []

    if len(lfns) == 1:
        try:
            with open(lfns[0], "r") as f:
                lfns = f.read().splitlines()
        except Exception:
            pass

    result = dirac.getLfnMetadata(lfns, printOutput=True)
    if not result["OK"]:
        print("ERROR: ", result["Message"])
        exitCode = 2

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :   dirac-admin-add-site
# Author : Federico Stagni
########################################################################
"""
Add a new DIRAC SiteName to DIRAC Configuration, including one or more CEs.
If site is already in the CS with another name, error message will be produced.
If site is already in the CS with the right name, only new CEs will be added.

Example:
  $ dirac-admin-add-site LCG.IN2P3.fr IN2P3-Site ce01.in2p3.fr
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC import exit as DIRACExit, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getDIRACSiteName
from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(
        "DIRACSiteName: Name of the site for DIRAC in the form GRID.LOCATION.COUNTRY " "(ie: LCG.CERN.ch)"
    )
    Script.registerArgument("GridSiteName:  Name of the site in the Grid (ie: CERN-PROD)")
    Script.registerArgument(["CE:           Name of the CE to be included in the site (ie: ce111.cern.ch)"])
    Script.parseCommandLine(ignoreErrors=True)

    diracSiteName, gridSiteName, ces = Script.getPositionalArgs(group=True)

    try:
        diracGridType, place, country = diracSiteName.split(".")
    except ValueError:
        gLogger.error("The DIRACSiteName should be of the form GRID.LOCATION.COUNTRY for example LCG.CERN.ch")
        DIRACExit(-1)

    result = getDIRACSiteName(gridSiteName)
    newSite = True
    if result["OK"] and result["Value"]:
        if len(result["Value"]) > 1:
            gLogger.notice("%s GOC site name is associated with several DIRAC sites:" % gridSiteName)
            for i, dSite in enumerate(result["Value"]):
                gLogger.notice("%d: %s" % (i, dSite))
            inp = six.moves.input("Enter your choice number: ")
            try:
                inp = int(inp)
            except ValueError:
                gLogger.error("You should enter an integer number")
                DIRACExit(-1)
            if 0 <= inp < len(result["Value"]):
                diracCSSite = result["Value"][inp]
            else:
                gLogger.error("Number out of range: %d" % inp)
                DIRACExit(-1)
        else:
            diracCSSite = result["Value"][0]
        if diracCSSite == diracSiteName:
            gLogger.notice("Site with GOC name %s is already defined as %s" % (gridSiteName, diracSiteName))
            newSite = False
        else:
            gLogger.error("ERROR: Site with GOC name %s is already defined as %s" % (gridSiteName, diracCSSite))
            DIRACExit(-1)
    else:
        gLogger.error("ERROR getting DIRAC site name of %s" % gridSiteName, result.get("Message"))

    csAPI = CSAPI()

    if newSite:
        gLogger.notice("Site to CS: %s" % diracSiteName)
        res = csAPI.addSite(diracSiteName, {"Name": gridSiteName})
        if not res["OK"]:
            gLogger.error("Failed adding site to CS", res["Message"])
            DIRACExit(1)
        res = csAPI.commit()
        if not res["OK"]:
            gLogger.error("Failure committing to CS", res["Message"])
            DIRACExit(3)

    for ce in ces:
        gLogger.notice("Adding CE %s" % ce)
        res = csAPI.addCEtoSite(diracSiteName, ce)
        if not res["OK"]:
            gLogger.error("Failed adding CE %s to CS" % ce, res["Message"])
            DIRACExit(2)
        res = csAPI.commit()
        if not res["OK"]:
            gLogger.error("Failure committing to CS", res["Message"])
            DIRACExit(3)


if __name__ == "__main__":
    main()



""" ResourceManagementDB:
    This module provides definition of the DB tables, and methods to access them.

    Written using sqlalchemy declarative_base



    For extending the ResourceStatusDB tables:

    1) In the extended module, call:

    from DIRAC.ResourceStatusSystem.DB.ResourceManagementDB import rmsBase, TABLESLIST
    TABLESLIST = TABLESLIST + [list of new table names]

    2) provide a declarative_base definition of the tables (new or extended) in the extension module

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import datetime
from sqlalchemy.orm import class_mapper
from sqlalchemy.orm.query import Query
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, String, DateTime, exc, Text, Integer, Float

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.SQLAlchemyDB import SQLAlchemyDB
from DIRAC.ResourceStatusSystem.Utilities import Utils


# Defining the tables

TABLESLIST = [
    "AccountingCache",
    "DowntimeCache",
    "GGUSTicketsCache",
    "JobCache",
    "PilotCache",
    "PolicyResult",
    "SpaceTokenOccupancyCache",
    "TransferCache",
]

rmsBase = declarative_base()


class AccountingCache(rmsBase):
    """AccountingCache table"""

    __tablename__ = "AccountingCache"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    name = Column("Name", String(64), nullable=False, primary_key=True)
    plotname = Column("PlotName", String(64), nullable=False, primary_key=True)
    plottype = Column("PlotType", String(16), nullable=False, primary_key=True)
    lastchecktime = Column("LastCheckTime", DateTime, nullable=False)
    result = Column("Result", Text, nullable=False)
    dateeffective = Column("DateEffective", DateTime, nullable=False)

    def fromDict(self, dictionary):
        """
        Fill the fields of the AccountingCache object from a dictionary

        :param dictionary: Dictionary to fill a single line
        :type arguments: dict
        """

        self.name = dictionary.get("Name", self.name)
        self.plotname = dictionary.get("PlotName", self.plotname)
        self.plottype = dictionary.get("PlotType", self.plottype)
        self.lastchecktime = dictionary.get(
            "LastCheckTime",
            self.lastchecktime.replace(microsecond=0)
            if self.lastchecktime
            else datetime.datetime.utcnow().replace(microsecond=0),
        )
        self.result = dictionary.get("Result", self.result)
        self.dateeffective = dictionary.get(
            "DateEffective",
            self.dateeffective.replace(microsecond=0)
            if self.dateeffective
            else datetime.datetime.utcnow().replace(microsecond=0),
        )

    def toList(self):
        """Simply returns a list of column values"""
        return [self.name, self.plotname, self.plottype, self.lastchecktime, self.result, self.dateeffective]


class DowntimeCache(rmsBase):
    """DowntimeCache table"""

    __tablename__ = "DowntimeCache"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    downtimeid = Column("DowntimeID", String(127), nullable=False, primary_key=True)
    name = Column("Name", String(64), nullable=False)
    element = Column("Element", String(32), nullable=False)
    gocdbservicetype = Column("GOCDBServiceType", String(32), nullable=True)
    severity = Column("Severity", String(32), nullable=False)
    description = Column("Description", String(512), nullable=False)
    link = Column("Link", String(255), nullable=True)
    startdate = Column("StartDate", DateTime, nullable=False)
    enddate = Column("EndDate", DateTime, nullable=False)
    dateeffective = Column("DateEffective", DateTime, nullable=False)
    lastchecktime = Column("LastCheckTime", DateTime, nullable=False)

    def fromDict(self, dictionary):
        """
        Fill the fields of the DowntimeCache object from a dictionary
        """

        self.downtimeid = dictionary.get("DowntimeID", self.downtimeid)
        self.name = dictionary.get("Name", self.name)
        self.element = dictionary.get("Element", self.element)
        self.gocdbservicetype = dictionary.get("GOCDBServiceType", self.gocdbservicetype)
        self.severity = dictionary.get("Severity", self.severity)
        self.description = dictionary.get("Description", self.description)
        self.link = dictionary.get("Link", self.link)
        self.startdate = dictionary.get("StartDate", self.startdate)
        self.enddate = dictionary.get("EndDate", self.enddate)
        self.dateeffective = dictionary.get(
            "DateEffective",
            self.dateeffective.replace(microsecond=0)
            if self.dateeffective
            else datetime.datetime.utcnow().replace(microsecond=0),
        )
        self.lastchecktime = dictionary.get(
            "LastCheckTime",
            self.lastchecktime.replace(microsecond=0)
            if self.lastchecktime
            else datetime.datetime.utcnow().replace(microsecond=0),
        )

    def toList(self):
        """Simply returns a list of column values"""
        return [
            self.downtimeid,
            self.name,
            self.element,
            self.gocdbservicetype,
            self.severity,
            self.description,
            self.link,
            self.startdate,
            self.enddate,
            self.dateeffective,
            self.lastchecktime,
        ]


class GGUSTicketsCache(rmsBase):
    """GGUSTicketsCache table"""

    __tablename__ = "GGUSTicketsCache"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    gocsite = Column("GocSite", String(64), nullable=False, primary_key=True)
    tickets = Column("Tickets", String(1024), nullable=False)
    opentickets = Column("OpenTickets", Integer, nullable=False, server_default="0")
    link = Column("Link", String(1024), nullable=False)
    lastchecktime = Column("LastCheckTime", DateTime, nullable=False)

    def fromDict(self, dictionary):
        """
        Fill the fields of the GGUSTicketsCache object from a dictionary

        :param dictionary: Dictionary to fill a single line
        :type arguments: dict
        """

        self.tickets = dictionary.get("Tickets", self.tickets)
        self.opentickets = dictionary.get("OpenTickets", self.opentickets)
        self.gocsite = dictionary.get("GocSite", self.gocsite)
        self.link = dictionary.get("Link", self.link)
        self.lastchecktime = dictionary.get(
            "LastCheckTime",
            self.lastchecktime.replace(microsecond=0)
            if self.lastchecktime
            else datetime.datetime.utcnow().replace(microsecond=0),
        )

    def toList(self):
        """Simply returns a list of column values"""
        return [self.gocsite, self.tickets, self.opentickets, self.link, self.lastchecktime]


class JobCache(rmsBase):
    """JobCache table"""

    __tablename__ = "JobCache"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    site = Column("Site", String(64), nullable=False, primary_key=True)
    status = Column("Status", String(16), nullable=False)
    efficiency = Column("Efficiency", Float(asdecimal=False), nullable=False, server_default="0")
    maskstatus = Column("MaskStatus", String(32), nullable=False)
    lastchecktime = Column("LastCheckTime", DateTime, nullable=False)

    def fromDict(self, dictionary):
        """
        Fill the fields of the JobCache object from a dictionary
        """

        self.site = dictionary.get("Site", self.site)
        self.status = dictionary.get("Status", self.status)
        self.efficiency = dictionary.get("Efficiency", self.efficiency)
        self.maskstatus = dictionary.get("MaskStatus", self.maskstatus)
        self.lastchecktime = dictionary.get(
            "LastCheckTime",
            self.lastchecktime.replace(microsecond=0)
            if self.lastchecktime
            else datetime.datetime.utcnow().replace(microsecond=0),
        )

    def toList(self):
        """Simply returns a list of column values"""
        return [self.site, self.status, self.efficiency, self.maskstatus, self.lastchecktime]


class PilotCache(rmsBase):
    """PilotCache table"""

    __tablename__ = "PilotCache"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    site = Column("Site", String(64), nullable=False, primary_key=True)
    ce = Column("CE", String(64), nullable=False, primary_key=True)
    vo = Column("VO", String(64), nullable=False, primary_key=True, server_default="all")
    status = Column("Status", String(16), nullable=False)
    pilotjobeff = Column("PilotJobEff", Float(asdecimal=False), nullable=False, server_default="0")
    pilotsperjob = Column("PilotsPerJob", Float(asdecimal=False), nullable=False, server_default="0")
    lastchecktime = Column("LastCheckTime", DateTime, nullable=False)

    columnsOrder = ["Site", "CE", "Status", "PilotJobEff", "PilotsPerJob", "LastCheckTime", "VO"]

    def fromDict(self, dictionary):
        """
        Fill the fields of the PilotCache object from a dictionary

        :param dictionary: Dictionary to fill a single line
        :type arguments: dict
        """

        self.site = dictionary.get("Site", self.site)
        self.ce = dictionary.get("CE", self.ce)
        self.vo = dictionary.get("VO", self.vo)
        self.status = dictionary.get("Status", self.status)
        self.pilotjobeff = dictionary.get("PilotJobEff", self.pilotjobeff)
        self.pilotsperjob = dictionary.get("PilotsPerJob", self.pilotsperjob)
        self.lastchecktime = dictionary.get(
            "LastCheckTime",
            self.lastchecktime.replace(microsecond=0)
            if self.lastchecktime
            else datetime.datetime.utcnow().replace(microsecond=0),
        )

    def toList(self):
        """Simply returns a list of column values"""
        return [self.site, self.ce, self.status, self.pilotjobeff, self.pilotsperjob, self.lastchecktime, self.vo]


class PolicyResult(rmsBase):
    """PolicyResult table"""

    __tablename__ = "PolicyResult"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    policyname = Column("PolicyName", String(64), nullable=False, primary_key=True)
    statustype = Column("StatusType", String(16), nullable=False, server_default="", primary_key=True)
    element = Column("Element", String(32), nullable=False, primary_key=True)
    name = Column("Name", String(64), nullable=False, primary_key=True)
    vo = Column("VO", String(64), nullable=False, primary_key=True, server_default="all")
    status = Column("Status", String(16), nullable=False)
    reason = Column("Reason", String(512), nullable=False, server_default="Unspecified")
    dateeffective = Column("DateEffective", DateTime, nullable=False)
    lastchecktime = Column("LastCheckTime", DateTime, nullable=False)

    def fromDict(self, dictionary):
        """
        Fill the fields of the PolicyResult object from a dictionary

        :param dictionary: Dictionary to fill a single line
        :type arguments: dict
        """

        self.policyname = dictionary.get("PolicyName", self.policyname)
        self.statustype = dictionary.get("StatusType", self.statustype)
        self.element = dictionary.get("Element", self.element)
        self.name = dictionary.get("Name", self.name)
        self.vo = dictionary.get("VO", self.vo)
        self.status = dictionary.get("Status", self.status)
        self.reason = dictionary.get("Reason", self.reason)
        self.dateeffective = dictionary.get(
            "DateEffective",
            self.dateeffective.replace(microsecond=0)
            if self.dateeffective
            else datetime.datetime.utcnow().replace(microsecond=0),
        )
        self.lastchecktime = dictionary.get(
            "LastCheckTime",
            self.lastchecktime.replace(microsecond=0)
            if self.lastchecktime
            else datetime.datetime.utcnow().replace(microsecond=0),
        )

    def toList(self):
        """Simply returns a list of column values"""
        return [
            self.policyname,
            self.statustype,
            self.element,
            self.name,
            self.status,
            self.reason,
            self.dateeffective,
            self.lastchecktime,
            self.vo,
        ]


class SpaceTokenOccupancyCache(rmsBase):
    """SpaceTokenOccupancyCache table"""

    __tablename__ = "SpaceTokenOccupancyCache"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    endpoint = Column("Endpoint", String(128), nullable=False, primary_key=True)
    token = Column("Token", String(64), nullable=False, primary_key=True)
    guaranteed = Column("Guaranteed", Float(asdecimal=False), nullable=False, server_default="0")
    free = Column("Free", Float(asdecimal=False), nullable=False, server_default="0")
    total = Column("Total", Float(asdecimal=False), nullable=False, server_default="0")
    lastchecktime = Column("LastCheckTime", DateTime, nullable=False)

    def fromDict(self, dictionary):
        """
        Fill the fields of the SpaceTokenOccupancyCache object from a dictionary

        :param dictionary: Dictionary to fill a single line
        :type arguments: dict
        """

        self.endpoint = dictionary.get("Endpoint", self.endpoint)
        self.token = dictionary.get("Token", self.token)
        self.guaranteed = dictionary.get("Guaranteed", self.guaranteed)
        self.free = dictionary.get("Free", self.free)
        self.total = dictionary.get("Total", self.total)
        self.lastchecktime = dictionary.get(
            "LastCheckTime",
            self.lastchecktime.replace(microsecond=0)
            if self.lastchecktime
            else datetime.datetime.utcnow().replace(microsecond=0),
        )

    def toList(self):
        """Simply returns a list of column values"""
        return [self.endpoint, self.token, self.guaranteed, self.free, self.total, self.lastchecktime]


class TransferCache(rmsBase):
    """TransferCache table"""

    __tablename__ = "TransferCache"
    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    sourcename = Column("SourceName", String(64), nullable=False, primary_key=True)
    destinationname = Column("DestinationName", String(64), nullable=False, primary_key=True)
    metric = Column("Metric", String(16), nullable=False, primary_key=True)
    value = Column("Value", Float(asdecimal=False), nullable=False, server_default="0")
    lastchecktime = Column("LastCheckTime", DateTime, nullable=False)

    def fromDict(self, dictionary):
        """
        Fill the fields of the TransferCache object from a dictionary

        :param dictionary: Dictionary to fill a single line
        :type arguments: dict
        """

        self.sourcename = dictionary.get("SourceName", self.sourcename)
        self.destinationname = dictionary.get("DestinationName", self.destinationname)
        self.metric = dictionary.get("Metric", self.metric)
        self.value = dictionary.get("Value", self.value)
        self.lastchecktime = dictionary.get(
            "LastCheckTime",
            self.lastchecktime.replace(microsecond=0)
            if self.lastchecktime
            else datetime.datetime.utcnow().replace(microsecond=0),
        )

    def toList(self):
        """Simply returns a list of column values"""
        return [self.sourcename, self.destinationname, self.metric, self.value, self.lastchecktime]


class ResourceManagementDB(SQLAlchemyDB):
    """
    Class that defines the methods to interact to the ResourceManagementDB tables
    """

    def __init__(self):
        """c'tor

        :param self: self reference
        """

        super(ResourceManagementDB, self).__init__()

        # This is the list of tables that will be created.
        # It can be extended in an extension module
        self.tablesList = getattr(Utils.voimport("DIRAC.ResourceStatusSystem.DB.ResourceManagementDB"), "TABLESLIST")
        self._initializeConnection("ResourceStatus/ResourceManagementDB")

        # Create required tables
        self._createTablesIfNotThere(self.tablesList)

    # Extended SQL methods ######################################################

    def addOrModify(self, table, params):
        """
        Using the PrimaryKeys of the table, it looks for the record in the database.
        If it is there, it is updated, if not, it is inserted as a new entry.

        :param table: table where to add or modify
        :type table: str
        :param params: dictionary of what to add or modify
        :type params: dict

        :return: S_OK() || S_ERROR()
        """

        session = self.sessionMaker_o()

        found = False
        for ext in self.extensions:
            try:
                table_c = getattr(__import__(ext + __name__, globals(), locals(), [table]), table)
                found = True
                break
            except (ImportError, AttributeError):
                continue
        # If not found in extensions, import it from DIRAC base (this same module).
        if not found:
            table_c = getattr(__import__(__name__, globals(), locals(), [table]), table)

        columns = [key.name for key in class_mapper(table_c).columns]
        primaryKeys = [key.name for key in class_mapper(table_c).primary_key]

        try:
            select = Query(table_c, session=session)
            for columnName, columnValue in params.items():
                if not columnValue or columnName not in primaryKeys:
                    continue
                column_a = getattr(table_c, columnName.lower())
                if isinstance(columnValue, (list, tuple)):
                    select = select.filter(column_a.in_(list(columnValue)))
                elif isinstance(columnValue, six.string_types):
                    select = select.filter(column_a == columnValue)
                else:
                    self.log.error("type(columnValue) == %s" % type(columnValue))

            res = select.first()  # the selection is done via primaryKeys only
            if not res:  # if not there, let's insert it
                return self.insert(table, params)

            # Treating case of time value updates
            if "LastCheckTime" in columns and not params.get("LastCheckTime"):
                params["LastCheckTime"] = None
            if "DateEffective" in columns and not params.get("DateEffective"):
                params["DateEffective"] = None

            # now we assume we need to modify
            for columnName, columnValue in params.items():
                if columnName == "LastCheckTime" and not columnValue:  # we always update lastCheckTime
                    columnValue = datetime.datetime.utcnow().replace(microsecond=0)
                if columnName == "DateEffective" and not columnValue:  # we always update DateEffective, if there
                    columnValue = datetime.datetime.utcnow().replace(microsecond=0)
                if columnValue:
                    setattr(res, columnName.lower(), columnValue)

            session.commit()
            return S_OK()

        except exc.SQLAlchemyError as e:
            session.rollback()
            self.log.exception("addOrModify: unexpected exception", lException=e)
            return S_ERROR("addOrModify: unexpected exception %s" % e)
        finally:
            session.close()



""" Test class for agents
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

import pytest
from mock import MagicMock

from DIRAC import gLogger

# sut
from DIRAC.ResourceStatusSystem.Agent.SummarizeLogsAgent import SummarizeLogsAgent

# mocks
rsClientMock = MagicMock()

# useful stuff
sOk = {"OK": True, "Value": []}
sError = {"OK": False, "Message": "a mess"}


columns = ["ID", "Name", "StatusType", "Status", "ElementType", "TokenOwner"]
el101 = [101, "RRCKI-BUFFER", "WriteAccess", "Active", "StorageElement", "rs_svc"]
el102 = [102, "RRCKI-BUFFER", "WriteAccess", "Active", "StorageElement", "rs_svc"]
el103 = [103, "RRCKI-BUFFER", "ReadAccess", "Active", "StorageElement", "rs_svc"]
el104 = [104, "RRCKI-BUFFER", "ReadAccess", "Active", "StorageElement", "rs_svc"]
el105 = [105, "RRCKI-BUFFER", "WriteAccess", "Banned", "StorageElement", "rs_svc"]

response1 = {"OK": True, "Value": [el101], "Columns": columns}

response2 = {"OK": True, "Value": [el101, el102], "Columns": columns}

response3 = {"OK": True, "Value": [el101, el102, el103], "Columns": columns}

response4 = {"OK": True, "Value": [el101, el102, el103, el104], "Columns": columns}

response4_mixed = {"OK": True, "Value": [el102, el103, el104, el101], "Columns": columns}

response5 = {"OK": True, "Value": [el101, el102, el103, el104, el105], "Columns": columns}

mockAM = MagicMock()


@pytest.fixture()
def sla(mocker):
    mocker.patch("DIRAC.ResourceStatusSystem.Agent.SummarizeLogsAgent.AgentModule", new=mockAM)
    theSLA = SummarizeLogsAgent()
    theSLA.log = gLogger
    theSLA.log.setLevel("DEBUG")
    theSLA.am_getOption = mockAM
    theSLA.rsClient = rsClientMock
    return theSLA


@pytest.mark.parametrize(
    "rsClientMockSelectStatusElementReturnValue, expected, expectedValue",
    [
        (sError, False, None),
        (sOk, True, (None, {})),
        ({"OK": True, "Value": [], "Columns": []}, True, (None, {})),
        (response1, True, (101, {("RRCKI-BUFFER", "WriteAccess"): [dict(zip(columns, el101))]})),
        (response2, True, (102, {("RRCKI-BUFFER", "WriteAccess"): [dict(zip(columns, el101))]})),
        (
            response3,
            True,
            (
                103,
                {
                    ("RRCKI-BUFFER", "WriteAccess"): [dict(zip(columns, el101))],
                    ("RRCKI-BUFFER", "ReadAccess"): [dict(zip(columns, el103))],
                },
            ),
        ),
        (
            response4,
            True,
            (
                104,
                {
                    ("RRCKI-BUFFER", "WriteAccess"): [dict(zip(columns, el101))],
                    ("RRCKI-BUFFER", "ReadAccess"): [dict(zip(columns, el103))],
                },
            ),
        ),
        (
            response4_mixed,
            True,
            (
                101,
                {
                    ("RRCKI-BUFFER", "WriteAccess"): [dict(zip(columns, el102))],
                    ("RRCKI-BUFFER", "ReadAccess"): [dict(zip(columns, el103))],
                },
            ),
        ),
        (
            response5,
            True,
            (
                105,
                {
                    ("RRCKI-BUFFER", "WriteAccess"): [dict(zip(columns, el101)), dict(zip(columns, el105))],
                    ("RRCKI-BUFFER", "ReadAccess"): [dict(zip(columns, el103))],
                },
            ),
        ),
    ],
)
def test__summarizeLogs(rsClientMockSelectStatusElementReturnValue, expected, expectedValue, sla):
    rsClientMock.selectStatusElement.return_value = rsClientMockSelectStatusElementReturnValue
    res = sla._summarizeLogs("element")
    assert res["OK"] == expected
    assert res.get("Value") == expectedValue


@pytest.mark.parametrize(
    "key, logs, rsClientMockSelectStatusElementReturnValue, expected",
    [
        ((None, None), [], sError, True),
        ((None, None), ["bof"], sError, False),
        ((None, None), [], sOk, True),
        ((None, None), [], {"OK": True, "Value": [], "Columns": ["Status", "TokenOwner"]}, True),
        ((None, None), [], {"OK": True, "Value": [["Active", "rs_svc"]], "Columns": ["status", "tokenowner"]}, True),
        (
            ("RRCKI-BUFFER", "WriteAccess"),  # this should remove everything
            [dict(zip(columns, el101))],
            {"OK": True, "Value": [["Active", "rs_svc"]], "Columns": ["status", "tokenowner"]},
            True,
        ),
        (
            ("RRCKI-BUFFER", "WriteAccess"),  # this should keep it
            [dict(zip(columns, el101))],
            {"OK": True, "Value": [["Banned", "rs_svc"]], "Columns": ["status", "tokenowner"]},
            True,
        ),
    ],
)
def test__registerLogs(key, logs, rsClientMockSelectStatusElementReturnValue, expected, sla):
    rsClientMock.selectStatusElement.return_value = rsClientMockSelectStatusElementReturnValue
    res = sla._registerLogs("Resource", key, logs)
    assert res["OK"] == expected



""" StorageOccupancy records the Storage Elements occupancy over time
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.AccountingSystem.Client.Types.BaseAccountingType import BaseAccountingType


class StorageOccupancy(BaseAccountingType):
    """StorageOccupancy as extension of BaseAccountingType.

    It is filled by the RSS Command FreeDiskSpace every time the
    command is executed (from Agent CacheFeederAgent)
    """

    def __init__(self):
        """constructor."""
        super(StorageOccupancy, self).__init__()

        self.definitionKeyFields = [
            ("Site", "VARCHAR(64)"),
            ("Endpoint", "VARCHAR(255)"),
            ("StorageElement", "VARCHAR(64)"),
            ("SpaceType", "VARCHAR(64)"),
        ]  # (Total, Free, Used)

        self.definitionAccountingFields = [("Space", "BIGINT UNSIGNED")]

        self.bucketsLength = [
            (86400 * 2, 3600),  # <2d = 1h
            (86400 * 10, 3600 * 6),  # <10d = 6h
            (86400 * 40, 3600 * 12),  # <40d = 12h
            (86400 * 30 * 6, 86400 * 2),  # <6m = 2d
            (86400 * 600, 86400 * 7),  # >6m = 1w
        ]

        self.checkType()



#!/usr/bin/env python

"""
Ban one or more Storage Elements for usage

Example:
  $ dirac-admin-ban-se M3PEC-disk
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    read = True
    write = True
    check = True
    remove = True
    sites = []
    mute = False

    Script.registerSwitch("r", "BanRead", "     Ban only reading from the storage element")
    Script.registerSwitch("w", "BanWrite", "     Ban writing to the storage element")
    Script.registerSwitch("k", "BanCheck", "     Ban check access to the storage element")
    Script.registerSwitch("v", "BanRemove", "    Ban remove access to the storage element")
    Script.registerSwitch("a", "All", "    Ban all access to the storage element")
    Script.registerSwitch("m", "Mute", "     Do not send email")
    Script.registerSwitch(
        "S:", "Site=", "     Ban all SEs associate to site (note that if writing is allowed, check is always allowed)"
    )
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["seGroupList: list of SEs or comma-separated SEs"])

    switches, ses = Script.parseCommandLine(ignoreErrors=True)

    for switch in switches:
        if switch[0].lower() in ("r", "banread"):
            write = False
            check = False
            remove = False
        if switch[0].lower() in ("w", "banwrite"):
            read = False
            check = False
            remove = False
        if switch[0].lower() in ("k", "bancheck"):
            read = False
            write = False
            remove = False
        if switch[0].lower() in ("v", "banremove"):
            read = False
            write = False
            check = False
        if switch[0].lower() in ("a", "all"):
            pass
        if switch[0].lower() in ("m", "mute"):
            mute = True
        if switch[0].lower() in ("s", "site"):
            sites = switch[1].split(",")

    # from DIRAC.ConfigurationSystem.Client.CSAPI           import CSAPI
    from DIRAC import gConfig, gLogger
    from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
    from DIRAC.Core.Security.ProxyInfo import getProxyInfo
    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin
    from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus
    from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup, DMSHelpers

    ses = resolveSEGroup(ses)
    diracAdmin = DiracAdmin()
    setup = gConfig.getValue("/DIRAC/Setup", "")
    if not setup:
        print("ERROR: Could not contact Configuration Service")
        DIRAC.exit(2)

    res = getProxyInfo()
    if not res["OK"]:
        gLogger.error("Failed to get proxy information", res["Message"])
        DIRAC.exit(2)

    userName = res["Value"].get("username")
    if not userName:
        gLogger.error("Failed to get username for proxy")
        DIRAC.exit(2)

    for site in sites:
        res = DMSHelpers().getSEsForSite(site)
        if not res["OK"]:
            gLogger.error(res["Message"], site)
            DIRAC.exit(-1)
        ses.extend(res["Value"])

    if not ses:
        gLogger.error("There were no SEs provided")
        DIRAC.exit(-1)

    readBanned = []
    writeBanned = []
    checkBanned = []
    removeBanned = []

    resourceStatus = ResourceStatus()

    res = resourceStatus.getElementStatus(ses, "StorageElement")
    if not res["OK"]:
        gLogger.error("Storage Element %s does not exist" % ses)
        DIRAC.exit(-1)

    reason = "Forced with dirac-admin-ban-se by %s" % userName

    for se, seOptions in res["Value"].items():

        resW = resC = resR = {"OK": False}

        # Eventually, we will get rid of the notion of InActive, as we always write Banned.
        if read and "ReadAccess" in seOptions:

            if seOptions["ReadAccess"] == "Banned":
                gLogger.notice("Read access already banned", se)
                resR["OK"] = True
            elif not seOptions["ReadAccess"] in ["Active", "Degraded", "Probing"]:
                gLogger.notice(
                    "Read option for %s is %s, instead of %s"
                    % (se, seOptions["ReadAccess"], ["Active", "Degraded", "Probing"])
                )
                gLogger.notice("Try specifying the command switches")
            else:

                resR = resourceStatus.setElementStatus(se, "StorageElement", "ReadAccess", "Banned", reason, userName)
                # res = csAPI.setOption( "%s/%s/ReadAccess" % ( storageCFGBase, se ), "InActive" )
                if not resR["OK"]:
                    gLogger.error("Failed to update %s read access to Banned" % se)
                else:
                    gLogger.notice("Successfully updated %s read access to Banned" % se)
                    readBanned.append(se)

        # Eventually, we will get rid of the notion of InActive, as we always write Banned.
        if write and "WriteAccess" in seOptions:

            if seOptions["WriteAccess"] == "Banned":
                gLogger.notice("Write access already banned", se)
                resW["OK"] = True
            elif not seOptions["WriteAccess"] in ["Active", "Degraded", "Probing"]:
                gLogger.notice(
                    "Write option for %s is %s, instead of %s"
                    % (se, seOptions["WriteAccess"], ["Active", "Degraded", "Probing"])
                )
                gLogger.notice("Try specifying the command switches")
            else:

                resW = resourceStatus.setElementStatus(se, "StorageElement", "WriteAccess", "Banned", reason, userName)
                # res = csAPI.setOption( "%s/%s/WriteAccess" % ( storageCFGBase, se ), "InActive" )
                if not resW["OK"]:
                    gLogger.error("Failed to update %s write access to Banned" % se)
                else:
                    gLogger.notice("Successfully updated %s write access to Banned" % se)
                    writeBanned.append(se)

        # Eventually, we will get rid of the notion of InActive, as we always write Banned.
        if check and "CheckAccess" in seOptions:

            if seOptions["CheckAccess"] == "Banned":
                gLogger.notice("Check access already banned", se)
                resC["OK"] = True
            elif not seOptions["CheckAccess"] in ["Active", "Degraded", "Probing"]:
                gLogger.notice(
                    "Check option for %s is %s, instead of %s"
                    % (se, seOptions["CheckAccess"], ["Active", "Degraded", "Probing"])
                )
                gLogger.notice("Try specifying the command switches")
            else:

                resC = resourceStatus.setElementStatus(se, "StorageElement", "CheckAccess", "Banned", reason, userName)
                # res = csAPI.setOption( "%s/%s/CheckAccess" % ( storageCFGBase, se ), "InActive" )
                if not resC["OK"]:
                    gLogger.error("Failed to update %s check access to Banned" % se)
                else:
                    gLogger.notice("Successfully updated %s check access to Banned" % se)
                    checkBanned.append(se)

        # Eventually, we will get rid of the notion of InActive, as we always write Banned.
        if remove and "RemoveAccess" in seOptions:

            if seOptions["RemoveAccess"] == "Banned":
                gLogger.notice("Remove access already banned", se)
                resC["OK"] = True
            elif not seOptions["RemoveAccess"] in ["Active", "Degraded", "Probing"]:
                gLogger.notice(
                    "Remove option for %s is %s, instead of %s"
                    % (se, seOptions["RemoveAccess"], ["Active", "Degraded", "Probing"])
                )
                gLogger.notice("Try specifying the command switches")
            else:

                resC = resourceStatus.setElementStatus(se, "StorageElement", "RemoveAccess", "Banned", reason, userName)
                # res = csAPI.setOption( "%s/%s/CheckAccess" % ( storageCFGBase, se ), "InActive" )
                if not resC["OK"]:
                    gLogger.error("Failed to update %s remove access to Banned" % se)
                else:
                    gLogger.notice("Successfully updated %s remove access to Banned" % se)
                    removeBanned.append(se)

        if not (resR["OK"] or resW["OK"] or resC["OK"]):
            DIRAC.exit(-1)

    if not (writeBanned or readBanned or checkBanned or removeBanned):
        gLogger.notice("No storage elements were banned")
        DIRAC.exit(-1)

    if mute:
        gLogger.notice("Email is muted by script switch")
        DIRAC.exit(0)

    subject = "%s storage elements banned for use" % len(writeBanned + readBanned + checkBanned + removeBanned)
    addressPath = "EMail/Production"
    address = Operations().getValue(addressPath, "")

    body = ""
    if read:
        body = "%s\n\nThe following storage elements were banned for reading:" % body
        for se in readBanned:
            body = "%s\n%s" % (body, se)
    if write:
        body = "%s\n\nThe following storage elements were banned for writing:" % body
        for se in writeBanned:
            body = "%s\n%s" % (body, se)
    if check:
        body = "%s\n\nThe following storage elements were banned for check access:" % body
        for se in checkBanned:
            body = "%s\n%s" % (body, se)
    if remove:
        body = "%s\n\nThe following storage elements were banned for remove access:" % body
        for se in removeBanned:
            body = "%s\n%s" % (body, se)

    if not address:
        gLogger.notice("'%s' not defined in Operations, can not send Mail\n" % addressPath, body)
        DIRAC.exit(0)

    res = diracAdmin.sendMail(address, subject, body)
    gLogger.notice("Notifying %s" % address)
    if res["OK"]:
        gLogger.notice(res["Value"])
    else:
        gLogger.notice(res["Message"])
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-site-info
# Author :  Stuart Paterson
########################################################################
"""
Print Configuration information for a given Site

Example:
  $ dirac-admin-site-info LCG.IN2P3.fr
  {'CE': 'cclcgceli01.in2p3.fr, cclcgceli03.in2p3.fr, sbgce1.in2p3.fr',
   'Coordinates': '4.8655:45.7825',
   'Mail': 'grid.admin@cc.in2p3.fr',
   'MoUTierLevel': '1',
   'Name': 'IN2P3-CC',
   'SE': 'IN2P3-disk, DIRAC-USER'}
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["Site:     Name of the Site"])
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    errorList = []

    for site in args:

        result = diracAdmin.getSiteSection(site, printOutput=True)
        if not result["OK"]:
            errorList.append((site, result["Message"]))
            exitCode = 2

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" Operation handler to set the status for transformation files

TODO: Is there any good reason why this is just not using a forwardDISET instead?
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities import DEncode
from DIRAC.RequestManagementSystem.private.OperationHandlerBase import OperationHandlerBase
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

__RCSID__ = "$Id$"


class SetFileStatus(OperationHandlerBase):
    """
    .. class:: SetFileStatus

    SetFileStatus operation handler
    """

    def __init__(self, operation=None, csPath=None):
        """c'tor

        :param self: self reference
        :param Operation operation: Operation instance
        :param str csPath: CS path for this handler
        """
        OperationHandlerBase.__init__(self, operation, csPath)

    def __call__(self):
        """It expects to find the arguments for tc.setFileStatusForTransformation in operation.Arguments"""
        try:
            setFileStatusDict = DEncode.decode(self.operation.Arguments)[0]
            self.log.debug("decoded filStatusDict=%s" % str(setFileStatusDict))
        except ValueError as error:
            self.log.exception(error)
            self.operation.Error = str(error)
            self.operation.Status = "Failed"
            return S_ERROR(str(error))

        tc = TransformationClient()
        setStatus = tc.setFileStatusForTransformation(
            setFileStatusDict["transformation"], setFileStatusDict["statusDict"], setFileStatusDict["force"]
        )

        if not setStatus["OK"]:
            errorStr = "failed to change status: %s" % setStatus["Message"]
            self.operation.Error = errorStr
            self.log.warn(errorStr)
            return S_ERROR(self.operation.Error)

        else:
            self.operation.Status = "Done"
            return S_OK()



"""Test for JobReport"""
# pylint: disable=protected-access, missing-docstring, invalid-name

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from mock import MagicMock

# sut
from DIRAC.WorkloadManagementSystem.Client.JobReport import JobReport


def test_jobReport(mocker):
    mocker.patch("DIRAC.WorkloadManagementSystem.Client.JobStateUpdateClient", side_effect=MagicMock())

    jr = JobReport(123)
    res = jr.setJobStatus("Matched", "minor_matched", "app_matched", sendFlag=False)
    assert res["OK"]
    res = jr.setJobStatus("Running", "minor_running", "app_running", sendFlag=False)
    assert res["OK"]
    res = jr.setJobParameter("par_1", "value_1", sendFlag=False)
    assert res["OK"]
    res = jr.setJobParameter("par_2", "value_2", sendFlag=False)
    assert res["OK"]
    res = jr.setJobParameters([("par_3", "value_3"), ("par_4", "value_4")], sendFlag=False)
    print(jr.jobParameters)
    jr.dump()



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-pilot-summary
# Author :  Stuart Paterson
########################################################################
"""
Pilot summary

Example:
  $ dirac-admin-pilot-summary
  CE                Status  Count Status Count Status Count Status  Count Status    Count Status    Count Status  Count
  sbgce1.in2p3.fr   Done    31
  lpsc-ce.in2p3.fr  Done    111
  lyogrid2.in2p3.fr Done    81
  egee-ce.jusieu.fr Aborted 81    Done   18
  cclcgce3.in2p3.fr Done    275
  marce01.in2p3.fr  Done    156
  node7.datagrid.fr Done    75
  cclcgceli01.fr    Aborted 1     Done   235
  ce0.m3pec.u-bo.fr Done    63
  grive11.ibcp.fr   Aborted 3     Done   90
  lptace01.msfg.fr  Aborted 3     Done   3     Done   90
  ipnls1.in2p3.fr   Done    87
  Total             Aborted 89    Done   1423  Ready  0     Running 0     Scheduled 0     Submitted 0     Waiting 0
  lappce01.in2p3.fr Aborted 1     Done   111
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

# pylint: disable=wrong-import-position
import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()

    result = diracAdmin.getPilotSummary()
    if result["OK"]:
        DIRAC.exit(0)
    else:
        print(result["Message"])
        DIRAC.exit(2)


if __name__ == "__main__":
    main()



""" IdProvider base class for various identity providers
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import gLogger

__RCSID__ = "$Id$"


class IdProvider(object):
    def __init__(self, parameters=None):
        self.log = gLogger.getSubLogger(self.__class__.__name__)
        self.parameters = parameters

    def setParameters(self, parameters):
        self.parameters = parameters



"""
  This module defines the state machine for the Productions
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.StateMachine import State, StateMachine


class ProductionsStateMachine(StateMachine):
    """Production Management System implementation of the state machine"""

    def __init__(self, state):
        """c'tor
        Defines the state machine transactions
        """
        super(ProductionsStateMachine, self).__init__(state)

        # States transitions
        self.states = {
            "Cleaned": State(4),  # final state
            # The Cleaning State should be added later. This State implies that there is an agent checking if
            # all the associated transformations are cleaned and which automatically updates the State to
            # Cleaned. For the moment the transition is directly from Stopped to Cleaned.
            # 'Cleaning'   : State( 3, ['Cleaned'] ),
            "Stopped": State(3, ["Active", "Flush", "Cleaned"], defState="Active"),
            "Flush": State(2, ["Active", "Cleaned"], defState="Active"),
            "Active": State(1, ["Flush", "Stopped", "Cleaned"], defState="Flush"),
            # initial state
            "New": State(0, ["Active", "Cleaned"], defState="Active"),
        }



# TO-DO: to be moved to tests directory

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import cmd
import sys
import os.path

from DIRAC.DataManagementSystem.Client.CmdDirCompletion.AbstractFileSystem import UnixLikeFileSystem
from DIRAC.DataManagementSystem.Client.CmdDirCompletion.DirectoryCompletion import DirectoryCompletion


class DirCompletion(cmd.Cmd):

    ulfs = UnixLikeFileSystem()
    dc = DirectoryCompletion(ulfs)

    def do_exit(self, args):
        sys.exit(0)

    def _listdir(self, args):
        if os.path.isdir(args):
            return os.listdir(args)
        else:
            return [args]

    def _ls(self, args):
        try:
            return self._listdir(args)
        except:
            return []

    def do_ls(self, args):
        print()
        print(" ".join(self._ls(args)))

    def complete_ls(self, text, line, begidx, endidx):
        # print
        result = []

        cur_input_line = line.split()
        cur_path = "."
        if len(cur_input_line) == 2:
            cur_path = cur_input_line[1]
            # print "cur_path:", cur_path

        result = self.dc.parse_text_line(text, cur_path, os.getcwd())

        return result


if __name__ == "__main__":

    cli = DirCompletion()
    cli.cmdloop()



""" :mod: RequestTaskTests
    =======================

    .. module: RequestTaskTests
    :synopsis: test cases for RequestTask class

    test cases for RequestTask class
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id $"
# #
# @file RequestTaskTests.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/03/27 15:59:40
# @brief Definition of RequestTaskTests class.
# # imports
import unittest
import importlib
from mock import Mock, MagicMock

# # SUT
from DIRAC.RequestManagementSystem.private.RequestTask import RequestTask

# # request client
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient

ReqClient = Mock(spec=ReqClient)
# # from DIRAC
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation

########################################################################
class RequestTaskTests(unittest.TestCase):
    """
    .. class:: RequestTaskTests

    """

    def setUp(self):
        """test case set up"""
        self.handlersDict = {"ForwardDISET": "DIRAC/RequestManagementSystem/private/ForwardDISET"}
        self.req = Request()
        self.req.RequestName = "foobarbaz"
        self.req.OwnerGroup = "lhcb_user"
        self.req.OwnerDN = "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=chaen/CN=705305/CN=Christophe Haen"
        self.op = Operation({"Type": "ForwardDISET", "Arguments": "tts10:helloWorldee"})
        self.req.addOperation(self.op)
        self.task = None
        self.mockRC = MagicMock()

        self.mockObjectOps = MagicMock()
        self.mockObjectOps.getSections.return_value = {"OK": True, "Value": ["DataProcessing", "DataManager"]}
        self.mockObjectOps.getOptionsDict.return_value = {
            "OK": True,
            "Value": {"Group": "lhcb_user", "User": "fstagni"},
        }
        self.mockOps = MagicMock()
        self.mockOps.return_value = self.mockObjectOps

    def tearDown(self):
        """test case tear down"""
        del self.req
        del self.op
        del self.task

    def testAPI(self):
        """test API"""
        rt = importlib.import_module("DIRAC.RequestManagementSystem.private.RequestTask")
        rt.gMonitor = MagicMock()
        rt.Operations = self.mockOps
        rt.CS = MagicMock()

        self.task = RequestTask(
            self.req.toJSON()["Value"],
            self.handlersDict,
            "csPath",
            "RequestManagement/RequestExecutingAgent",
            requestClient=self.mockRC,
        )
        self.task.requestClient = Mock(return_value=Mock(spec=ReqClient))
        self.task.requestClient().updateRequest = Mock()
        self.task.requestClient().updateRequest.return_value = {"OK": True, "Value": None}
        ret = self.task()
        self.assertEqual(ret["OK"], True, "call failed")

        ret = self.task.setupProxy()
        print(ret)


# # tests execution
if __name__ == "__main__":
    testLoader = unittest.TestLoader()
    requestTaskTests = testLoader.loadTestsFromTestCase(RequestTaskTests)
    suite = unittest.TestSuite([requestTaskTests])
    unittest.TextTestRunner(verbosity=3).run(suite)



########################################################################
# File: AdlerTestCase.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2011/02/11 09:08:19
########################################################################

""" :mod: AdlerTestCase
    =======================

    .. module: AdlerTestCase
    :synopsis: test case for DIRAC.Core.Utilities.Adler module
    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    test case for DIRAC.Core.Utilities.Adler module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

##
# @file AdlerTestCase.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2011/02/11 09:08:37
# @brief Definition of AdlerTestCase class.

# imports
import os
import unittest
import string
import tempfile

# from zlib import adler32
import zlib

# from DIRAC
from DIRAC.Core.Utilities import Adler

__RCSID__ = "$Id$"

########################################################################


class AdlerTestCase(unittest.TestCase):

    """
    .. class:: AdlerTestCase
    test case for DIRAC.Core.Utilities.Adler module
    """

    def setUp(self):
        self.emptyAdler = hex(zlib.adler32(b"") & 0xFFFFFFFF)[2:]
        self.lettersAdler = hex(zlib.adler32(string.ascii_letters.encode()) & 0xFFFFFFFF)[2:]

    def testStringAdler(self):
        """stringAdler tests"""
        # no arguments supplied - TypeError
        try:
            Adler.stringAdler()  # pylint: disable=no-value-for-parameter
        except Exception as error:
            self.assertEqual(isinstance(error, TypeError), True)
        # wrong argument type
        self.assertEqual(Adler.stringAdler([]), False)
        # empty string
        self.assertEqual(int(Adler.stringAdler("")), int(self.emptyAdler))
        # all letters
        self.assertEqual(Adler.stringAdler(string.ascii_letters), self.lettersAdler)

    def testConversion(self):
        """intAdlerToHex and hexAdlerToInt tests"""
        # no arguments
        try:
            Adler.intAdlerToHex()  # pylint: disable=no-value-for-parameter
        except Exception as error:
            self.assertEqual(isinstance(error, TypeError), True)
        # wrong type of arg (should it really print out to stdout)
        self.assertEqual(Adler.intAdlerToHex("a"), False)
        # normal operation
        self.assertEqual(int(Adler.intAdlerToHex(1)), Adler.hexAdlerToInt(Adler.intAdlerToHex(1)))
        self.assertEqual(Adler.hexAdlerToInt("0x01"), int(Adler.intAdlerToHex(Adler.hexAdlerToInt("0x01"))))

    def testFileAdler(self):
        """fileAdler tests"""
        # no args
        try:
            Adler.fileAdler()  # pylint: disable=no-value-for-parameter
        except Exception as error:
            self.assertEqual(isinstance(error, TypeError), True)
        # read-protected file
        self.assertEqual(Adler.fileAdler("/root/.login"), False)
        # inexisting file
        self.assertEqual(Adler.fileAdler("Stone/Dead/Norwegian/Blue/Parrot/In/Camelot"), False)
        # normal operation
        fd, path = tempfile.mkstemp("_adler32", "norewgian_blue")
        self.assertEqual(int(Adler.fileAdler(path)), int(self.emptyAdler))
        os.write(fd, string.ascii_letters.encode())
        self.assertEqual(Adler.fileAdler(path), self.lettersAdler)

    def testCompareAdler(self):
        """compareAdler tests"""
        # same adlers
        self.assertEqual(Adler.compareAdler(Adler.stringAdler(""), Adler.stringAdler("")), True)
        # diff adlers
        self.assertEqual(Adler.compareAdler(Adler.stringAdler(""), Adler.stringAdler(string.ascii_letters)), False)


# test suite execution
if __name__ == "__main__":
    TESTLOADER = unittest.TestLoader()
    SUITE = TESTLOADER.loadTestsFromTestCase(AdlerTestCase)
    unittest.TextTestRunner(verbosity=3).run(SUITE)



""" DowntimeCommand module will look into GOC DB to find announced downtimes for RSS-managed sites and resources.
    If found, downtimes are added to the internal RSS cache using ResourceManagementClient.

    GOCDB downtimes that are modified or deleted are also synced.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from six.moves.urllib_error import URLError
import re

from datetime import datetime, timedelta
from operator import itemgetter

from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.Core.LCG.GOCDBClient import GOCDBClient
from DIRAC.Core.Utilities.SiteSEMapping import getSEHosts, getStorageElementsHosts
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgPath
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import (
    getFTS3Servers,
    getGOCSiteName,
    getGOCSites,
    getGOCFTSName,
    getCESiteMapping,
)
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.ResourceStatusSystem.Client.ResourceManagementClient import ResourceManagementClient
from DIRAC.ResourceStatusSystem.Command.Command import Command


class DowntimeCommand(Command):
    """
    Downtime "master" Command or removed DTs.
    """

    def __init__(self, args=None, clients=None):

        super(DowntimeCommand, self).__init__(args, clients)

        if "GOCDBClient" in self.apis:
            self.gClient = self.apis["GOCDBClient"]
        else:
            self.gClient = GOCDBClient()

        if "ResourceManagementClient" in self.apis:
            self.rmClient = self.apis["ResourceManagementClient"]
        else:
            self.rmClient = ResourceManagementClient()

    def _storeCommand(self, result):
        """
        Stores the results of doNew method on the database.
        """

        for dt in result:
            resQuery = self.rmClient.addOrModifyDowntimeCache(
                downtimeID=dt["DowntimeID"],
                element=dt["Element"],
                name=dt["Name"],
                startDate=dt["StartDate"],
                endDate=dt["EndDate"],
                severity=dt["Severity"],
                description=dt["Description"],
                link=dt["Link"],
                gOCDBServiceType=dt["gOCDBServiceType"],
            )
        return resQuery

    def _cleanCommand(self, element, elementNames):
        """
        Clear Cache from expired DT.
        """

        resQuery = []

        for elementName in elementNames:
            # get the list of all DTs stored in the cache
            result = self.rmClient.selectDowntimeCache(element=element, name=elementName)

            if not result["OK"]:
                return result

            uniformResult = [dict(zip(result["Columns"], res)) for res in result["Value"]]

            currentDate = datetime.utcnow()

            if not uniformResult:
                continue

            # get the list of all ongoing DTs from GocDB
            gDTLinkList = self.gClient.getCurrentDTLinkList()
            if not gDTLinkList["OK"]:
                return gDTLinkList

            for dt in uniformResult:
                # if DT expired or DT not in the list of current DTs, then we remove it from the cache
                if dt["EndDate"] < currentDate or dt["Link"] not in gDTLinkList["Value"]:
                    result = self.rmClient.deleteDowntimeCache(downtimeID=dt["DowntimeID"])
                    resQuery.append(result)

        return S_OK(resQuery)

    def _prepareCommand(self):
        """
        DowntimeCommand requires four arguments:
        - name : <str>
        - element : Site / Resource
        - elementType: <str>

        If the elements are Site(s), we need to get their GOCDB names. They may
        not have, so we ignore them if they do not have.
        """

        if "name" not in self.args:
            return S_ERROR('"name" not found in self.args')
        elementName = self.args["name"]

        if "element" not in self.args:
            return S_ERROR('"element" not found in self.args')
        element = self.args["element"]

        if "elementType" not in self.args:
            return S_ERROR('"elementType" not found in self.args')
        elementType = self.args["elementType"]

        if element not in ["Site", "Resource"]:
            return S_ERROR("element is neither Site nor Resource")

        hours = None
        if "hours" in self.args:
            hours = self.args["hours"]

        gOCDBServiceType = None

        # Transform DIRAC site names into GOCDB topics
        if element == "Site":

            gocSite = getGOCSiteName(elementName)
            if not gocSite["OK"]:  # The site is most probably is not a grid site - not an issue, of course
                pass  # so, elementName remains unchanged
            else:
                elementName = gocSite["Value"]

        # The DIRAC se names mean nothing on the grid, but their hosts do mean.
        elif elementType == "StorageElement":
            # for SRM and SRM only, we need to distinguish if it's tape or disk
            # if it's not SRM, then gOCDBServiceType will be None (and we'll use them all)
            try:
                se = StorageElement(elementName)
                seOptions = se.options
                seProtocols = set(se.localAccessProtocolList) | set(se.localWriteProtocolList)
            except AttributeError:  # Sometimes the SE can't be instantiated properly
                self.log.error("Failure instantiating StorageElement object", elementName)
                return S_ERROR("Failure instantiating StorageElement")
            if "SEType" in seOptions and "srm" in seProtocols:
                # Type should follow the convention TXDY
                seType = seOptions["SEType"]
                diskSE = re.search("D[1-9]", seType) is not None
                tapeSE = re.search("T[1-9]", seType) is not None
                if tapeSE:
                    gOCDBServiceType = "srm.nearline"
                elif diskSE:
                    gOCDBServiceType = "srm"

            res = getSEHosts(elementName)
            if not res["OK"]:
                return res
            seHosts = res["Value"]

            if not seHosts:
                return S_ERROR("No seHost(s) for %s" % elementName)
            elementName = seHosts  # in this case it will return a list, because there might be more than one host only

        elif elementType in ["FTS", "FTS3"]:
            gOCDBServiceType = "FTS"
            # WARNING: this method presupposes that the server is an FTS3 type
            gocSite = getGOCFTSName(elementName)
            if not gocSite["OK"]:
                self.log.warn("FTS not in Resources/FTSEndpoints/FTS3 ?", elementName)
            else:
                elementName = gocSite["Value"]

        elif elementType == "ComputingElement":
            res = getCESiteMapping(elementName)
            if not res["OK"]:
                return res
            siteName = res["Value"][elementName]
            ceType = gConfig.getValue(
                cfgPath("Resources", "Sites", siteName.split(".")[0], siteName, "CEs", elementName, "CEType")
            )
            if ceType == "HTCondorCE":
                gOCDBServiceType = "org.opensciencegrid.htcondorce"
            elif ceType == "ARC":
                gOCDBServiceType = "ARC-CE"

        return S_OK((element, elementName, hours, gOCDBServiceType))

    def doNew(self, masterParams=None):
        """
        Gets the parameters to run, either from the master method or from its
        own arguments.

        For every elementName, unless it is given a list, in which case it contacts
        the gocdb client. The server is not very stable, so in case of failure tries
        a second time.

        If there are downtimes, are recorded and then returned.
        """

        if masterParams is not None:
            element, elementNames = masterParams
            hours = 120
            elementName = None
            gOCDBServiceType = None

        else:
            params = self._prepareCommand()
            if not params["OK"]:
                return params
            element, elementName, hours, gOCDBServiceType = params["Value"]
            if not isinstance(elementName, list):
                elementNames = [elementName]
            else:
                elementNames = elementName

        # WARNING: checking all the DT that are ongoing or starting in given <hours> from now
        try:
            results = self.gClient.getStatus(element, name=elementNames, startingInHours=hours)
        except URLError:
            try:
                # Let's give it a second chance..
                results = self.gClient.getStatus(element, name=elementNames, startingInHours=hours)
            except URLError as e:
                return S_ERROR(e)

        if not results["OK"]:
            return results
        results = results["Value"]

        if results is None:  # no downtimes found
            return S_OK(None)

        # cleaning the Cache
        if elementNames:
            cleanRes = self._cleanCommand(element, elementNames)
            if not cleanRes["OK"]:
                return cleanRes

        uniformResult = []

        # Humanize the results into a dictionary, not the most optimal, but readable
        for downtime, downDic in results.items():  # can be an iterator

            dt = {}

            dt["Name"] = downDic.get("URL", downDic.get("HOSTNAME", downDic.get("SITENAME")))
            if not dt["Name"]:
                return S_ERROR("URL, SITENAME and HOSTNAME are missing from downtime dictionary")

            dt["gOCDBServiceType"] = downDic.get("SERVICE_TYPE")

            if dt["gOCDBServiceType"] and gOCDBServiceType:
                if gOCDBServiceType.lower() != downDic["SERVICE_TYPE"].lower():
                    self.log.warn(
                        "SERVICE_TYPE mismatch",
                        "between GOCDB (%s) and CS (%s) for %s"
                        % (downDic["SERVICE_TYPE"], gOCDBServiceType, dt["Name"]),
                    )

            dt["DowntimeID"] = downtime
            dt["Element"] = element
            dt["StartDate"] = downDic["FORMATED_START_DATE"]
            dt["EndDate"] = downDic["FORMATED_END_DATE"]
            dt["Severity"] = downDic["SEVERITY"]
            dt["Description"] = downDic["DESCRIPTION"].replace("'", "")
            dt["Link"] = downDic["GOCDB_PORTAL_URL"]

            uniformResult.append(dt)

        storeRes = self._storeCommand(uniformResult)
        if not storeRes["OK"]:
            return storeRes

        return S_OK()

    def doCache(self):
        """
        Method that reads the cache table and tries to read from it. It will
        return a list with one dictionary describing the DT if there are results.
        """

        params = self._prepareCommand()
        if not params["OK"]:
            return params
        element, elementName, hours, gOCDBServiceType = params["Value"]

        result = self.rmClient.selectDowntimeCache(element=element, name=elementName, gOCDBServiceType=gOCDBServiceType)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK()

        uniformResult = [dict(zip(result["Columns"], res)) for res in result["Value"]]

        # 'targetDate' can be either now or in some 'hours' from now
        targetDate = datetime.utcnow()

        # dtOverlapping is a buffer to assure only one dt is returned
        # when there are overlapping outage/warning dt for same element
        # on top of the buffer we put the most recent outages
        # while at the bottom the most recent warnings,
        # assumption: uniformResult list is already ordered by resource/site name, severity, startdate
        dtOverlapping = []

        if hours is not None:
            # IN THE FUTURE
            targetDate = targetDate + timedelta(hours=hours)
            # sorting by 'StartDate' b/c if we look for DTs in the future
            # then we are interested in the earliest DTs
            uniformResult.sort(key=itemgetter("Name", "Severity", "StartDate"))

            for dt in uniformResult:
                if (dt["StartDate"] < targetDate) and (dt["EndDate"] > targetDate):
                    # the list is already ordered in a way that outages come first over warnings
                    # and the earliest outages are on top of other outages and warnings
                    # while the earliest warnings are on top of the other warnings
                    # so what ever comes first in the list is also what we are looking for
                    dtOverlapping = [dt]
                    break
        else:
            # IN THE PRESENT
            # sorting by 'EndDate' b/c if we look for DTs in the present
            # then we are interested in those DTs that last longer
            uniformResult.sort(key=itemgetter("Name", "Severity", "EndDate"))

            for dt in uniformResult:
                if (dt["StartDate"] < targetDate) and (dt["EndDate"] > targetDate):
                    # if outage, we put it on top of the overlapping buffer
                    # i.e. the latest ending outage is on top
                    if dt["Severity"].upper() == "OUTAGE":
                        dtOverlapping = [dt] + dtOverlapping
                    # if warning, we put it at the bottom of the overlapping buffer
                    # i.e. the latest ending warning is at the bottom
                    elif dt["Severity"].upper() == "WARNING":
                        dtOverlapping.append(dt)

        if not dtOverlapping:
            return S_OK()

        dtTop = dtOverlapping[0]
        if dtTop["Severity"].upper() == "OUTAGE":
            return S_OK(dtTop)
        else:
            return S_OK(dtOverlapping[-1])

    def doMaster(self):
        """Master method, which looks little bit spaghetti code, sorry !
        - It gets all sites and transforms them into gocSites.
        - It gets all the storage elements and transforms them into their hosts
        - It gets the the CEs (FTS and file catalogs will come).
        """

        gocSites = getGOCSites()
        if not gocSites["OK"]:
            return gocSites
        gocSites = gocSites["Value"]

        sesHosts = getStorageElementsHosts()
        if not sesHosts["OK"]:
            return sesHosts
        sesHosts = sesHosts["Value"]

        resources = sesHosts if sesHosts else []

        ftsServer = getFTS3Servers(hostOnly=True)
        if ftsServer["OK"] and ftsServer["Value"]:
            resources.extend(ftsServer["Value"])

        # TODO: file catalogs need also to use their hosts

        # fc = CSHelpers.getFileCatalogs()
        # if fc[ 'OK' ]:
        #  resources = resources + fc[ 'Value' ]

        res = getCESiteMapping()
        if res["OK"] and res["Value"]:
            resources.extend(list(res["Value"]))

        self.log.verbose("Processing Sites", ", ".join(gocSites if gocSites else ["NONE"]))

        siteRes = self.doNew(("Site", gocSites))
        if not siteRes["OK"]:
            self.metrics["failed"].append(siteRes["Message"])

        self.log.verbose("Processing Resources", ", ".join(resources if resources else ["NONE"]))

        resourceRes = self.doNew(("Resource", resources))
        if not resourceRes["OK"]:
            self.metrics["failed"].append(resourceRes["Message"])

        return S_OK(self.metrics)



""" Unit tests for PathFinder only for functions that I added
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import mock
import pytest
import random

from DIRAC import S_OK, S_ERROR
import DIRAC.ConfigurationSystem.Client.LocalConfiguration


doc_description = """
This is an example of a typical team description.

With indents and explanatory inserts::

  * possible option
  * possible option

UnprocessedUnit:
  Some interesting information

"""
doc_usage = """
Usage:
  my-script [options] ...
"""
doc_gen_opts = """
General options:
  -o  --option <value>         : Option=value to add
  -s  --section <value>        : Set base section for relative parsed options
  -c  --cert <value>           : Use server certificate to connect to Core Services
  -d  --debug                  : Set debug mode (-ddd is extra debug)
  -   --cfg=                   : Load additional config file
  -   --autoreload             : Automatically restart if there's any change in the module
  -   --license                : Show DIRAC's LICENSE
  -h  --help                   : Shows this help
"""
doc_options = """
Options:
  -H  --host <value>  : host
  -V  --vo <value>    : vo
"""
doc_arguments = """
Arguments:
  info:  some info
"""
doc_example = """
Example:
  $ my-script -H dirac
  Hello dirac
"""

doc_blocks = [doc_usage, doc_options, doc_gen_opts, doc_arguments, doc_example]

opts = ()
args = []
output = []
result = None
exitCode = None


def mock_notice(*a):
    global output
    output.append(" ".join(a))


def mock_exit(c=0):
    global exitCode
    exitCode = c


def mock_gnu(*a):
    global opts, args
    return opts, args


def action(a):
    global result
    result = a if a else "no"
    return S_OK()


@pytest.fixture
def localCFG(monkeypatch):
    monkeypatch.setattr(DIRAC.ConfigurationSystem.Client.LocalConfiguration.DIRAC, "exit", mock_exit)
    monkeypatch.setattr(DIRAC.ConfigurationSystem.Client.LocalConfiguration.getopt, "gnu_getopt", mock_gnu)
    monkeypatch.setattr(DIRAC.ConfigurationSystem.Client.LocalConfiguration.gLogger, "notice", mock_notice)
    localCFG = DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration()
    # It's local test, do not contact Configuration Server
    localCFG.disableCS()
    return localCFG


@pytest.mark.parametrize("blocks_order", [doc_blocks for i in range(10) if not random.shuffle(doc_blocks)])
def test_script_head(localCFG, blocks_order):
    global output
    output = []
    localCFG.setUsageMessage("\n".join([doc_description, "\n".join(blocks_order)]))
    localCFG.showHelp()
    assert "\n".join(output) == "".join([doc_description, doc_usage, doc_gen_opts, doc_arguments, doc_example])


def test_register_options(localCFG):
    global output, opts, args, result
    args = []
    output = []
    localCFG.setUsageMessage("\n".join([doc_description, "\n".join(doc_blocks)]))
    localCFG.registerCmdOpt("O:", "Option=", "my option", action)
    localCFG.registerCmdOpt("A", "AnotherOpt", "my another option", action)
    localCFG.registerCmdOpt("N", "NoActionOp", "my noaction option")
    localCFG.showHelp()
    optBlock = "\nOptions:\n  %s" % "\n  ".join(
        [
            "-O  --Option <value>         : my option",
            "-A  --AnotherOpt             : my another option",
            "-N  --NoActionOp             : my noaction option",
        ]
    )
    assert optBlock in "\n".join(output)

    # Check action result
    for opts, fn_res in [
        ([("-O", "y")], "y"),
        ([("--Option", "y")], "y"),
        ([("-A", None)], "no"),
        ([("--AnotherOpt", None)], "no"),
        ([("-N", None)], None),
    ]:
        result = None
        localCFG.isParsed = False
        localCFG.initialized = False
        assert localCFG.loadUserData() == S_OK()
        assert result == fn_res
        assert localCFG.getUnprocessedSwitches() == ([("N", None)] if opts[0] == ("-N", None) else [])


sArg = ("SingArg", "SingArg: my single argument", "  SingArg:  my single argument")
fArg = (
    "<ThisArg|ThatArg>",
    ("ThisArg: my this argument", "ThatArg: my that argument"),
    "  ThisArg:  my this argument\n  ThatArg:  my that argument",
)
lArg = ("ListArg [ListArg]", ["ListArg: my list of arguments"], "  ListArg:  my list of arguments")
aValues = ["yes", "no"]

list_optional = (lArg, False, None, "defVal")
float_optional = (fArg, False, None, "defVal")
single_optional = (sArg, False, None, "defVal")
list_mandatory = (lArg, True, None, "defVal")
float_mandatory = (fArg, True, None, "defVal")
single_mandatory = (sArg, True, None, "defVal")
list_checkvalues = (lArg, True, aValues, "defVal")
float_checkvalues = (fArg, True, aValues, "defVal")
single_checkvalues = (sArg, True, aValues, "defVal")

params = [
    ([list_optional, float_optional, single_optional], []),
    ([float_optional, list_optional, single_optional], []),
    ([list_mandatory, list_mandatory], []),
    ([single_mandatory, single_optional, single_mandatory], []),
    (
        [single_mandatory, float_mandatory, list_optional],
        [
            (["a", "b", "c", "d"], ["a", "b", "c", "d"], ["a", "b", ["c", "d"]]),
            (["a", "b", "c"], ["a", "b", "c"], ["a", "b", ["c"]]),
            (["a", "b"], ["a", "b"], ["a", "b", "defVal"]),
            (["a"], [], ()),
        ],
    ),
    (
        [single_mandatory, float_mandatory, list_mandatory],
        [
            (["a", "b", "c", "d"], ["a", "b", "c", "d"], ["a", "b", ["c", "d"]]),
            (["a", "b", "c"], ["a", "b", "c"], ["a", "b", ["c"]]),
            (["a", "b"], [], ()),
        ],
    ),
    (
        [single_mandatory, list_mandatory, float_mandatory],
        [
            (["a", "b", "c", "d"], ["a", "b", "c", "d"], ["a", ["b", "c"], "d"]),
            (["a", "b", "c"], ["a", "b", "c"], ["a", ["b"], "c"]),
            (["a", "b"], [], ()),
        ],
    ),
]


@pytest.mark.parametrize("argsData, expected", params)
def test_register_arguments(localCFG, argsData, expected):
    global output, opts, args, result, exitCode
    output = []
    opts = ()
    exitCode = 0
    useBlock = " [options] ..."
    argBlock = "\nArguments:"

    if not expected:
        with pytest.raises(Exception):
            for arg, mandatory, values, default in argsData:
                localCFG.registerCmdArg(arg[1], mandatory, values, default)
    else:
        for arg, mandatory, values, default in argsData:
            useBlock += " " + arg[0]
            for a in arg[2].split("\n"):
                argBlock += "\n" + a
                if values:
                    argBlock += " [%s]" % ", ".join(values)
                if default:
                    argBlock += " [default: defVal]"
                if not mandatory:
                    argBlock += " (optional)"
            localCFG.registerCmdArg(arg[1], mandatory, values, default)

    if expected:
        localCFG.showHelp()

        assert argBlock in "\n".join(output)
        assert useBlock in "\n".join(output)

        # Check action result
        for args, getargs, getgroupargs in expected:
            localCFG.isParsed = False
            localCFG.initialized = False
            assert localCFG.loadUserData() == S_OK()
            if not getargs:
                localCFG.getPositionalArguments()
                assert exitCode == 1
            else:
                assert localCFG.getPositionalArguments() == getargs
                assert localCFG.getPositionalArguments(True) == getgroupargs



""" This module contains utilities for parsing extensions in general, but mostly the VOMS extensions.
It has been done based on the reading of the VOMS standard (https://www.ogf.org/documents/GFD.182.pdf)
and on the RFC 5755 (http://www.ietf.org/rfc/rfc5755.txt)

This module relies on definition of the RFC 3281, which is the predecessor of 5755, but it still
seems to work for what we are interested in.

To summarize, the attributes we are interested in are called CertificateAttributes, and are stored in proxy extensions.
The VOMS extension in a proxy is a Sequence of Sequence (??) of CertificateAttribute. One Sequence is due to the fact
that you can embed more than one VO CertificateAttribute in one proxy. The other one was acknowledge as a an error in
the formal description (an Errata will come)

This is now pure python, but it might be interesting to wrap the existing
C library (https://github.com/italiangrid/voms) instead...

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pyasn1.codec.der.decoder import decode as der_decode
from pyasn1.error import PyAsn1Error
from pyasn1.type import namedtype, univ, char as asn1char
from pyasn1_modules import rfc2459, rfc3281
from DIRAC.Core.Security.m2crypto import (
    VOMS_EXTENSION_OID,
    VOMS_FQANS_OID,
    DN_MAPPING,
    VOMS_TAGS_EXT_OID,
    DIRAC_GROUP_OID,
)


class _ACSequence(univ.SequenceOf):
    """This describe a sequence of AC, as per GFD 182"""

    componentType = rfc3281.AttributeCertificate()


class _ACSequenceOfSequence(univ.SequenceOf):
    """A Sequence of Sequence of AC. In contradiction to GFD formal description,
    but that's what it is
    """

    componentType = _ACSequence()


# The Tag structure is used to describe things like the nickname
# (See OGF 3.6.4)
# Similarely to above, there is an extra Sequence layer above the TagContainer
class _VOMSTag(univ.Sequence):
    """Tag as per GOF 182"""

    componentType = namedtype.NamedTypes(
        namedtype.NamedType("name", univ.OctetString()),
        namedtype.NamedType("value", univ.OctetString()),
        namedtype.NamedType("qualifier", univ.OctetString()),
    )


class _VOMSTags(univ.SequenceOf):
    """Sequence of VOMSTag as per GOF 182"""

    componentType = _VOMSTag()


class _TagList(univ.Sequence):
    """TagList as per GOF 182"""

    componentType = namedtype.NamedTypes(
        namedtype.NamedType("policyAuthority", rfc2459.GeneralNames()), namedtype.NamedType("tags", _VOMSTags())
    )


class _TagContainer(univ.SequenceOf):
    """TagContainer as per GOF 182"""

    componentType = _TagList()


class _TagContainers(univ.SequenceOf):
    """Sequence of TagContainer
    Note sure why it is here, but that's how it is encoded
    """

    componentType = _TagContainer()


def decodeDIRACGroup(m2cert):
    """Decode the content of the dirac group extension

    :param m2cert: M2crypto x509 object, a certificate

    :returns: the dirac group

    :raises: same as retrieveExtension
    """

    diracGroupOctetString = retrieveExtension(m2cert, DIRAC_GROUP_OID)
    diracGroupUTF8Str, _rest = der_decode(diracGroupOctetString, asn1Spec=asn1char.IA5String())

    return diracGroupUTF8Str.asOctets().decode()


def _decodeASN1String(rdnNameAttrValue):
    """Tries to decode a string encoded with the following type:
    * BMPString
    * IA5String
    * PrintableString
    * TeletexString
    * UTF8String

    Most of these types come from the definition of the issuer field in RFC3280:
    * The basic attributes, defined as DirectoryString (4.1.2.4  Issuer)
    * the optional attributes (Appendix A.  Psuedo-ASN.1 Structures and OIDs)

    This utility function is needed for 2 reasons:
    * Not all the attributes are encoded the same way, and as we do not want to bother
      with zillions of `if` conditions, we may just as well try
    * It seems like the RFCs are not always respected, and the encoding is not always the correct one

    http://www.oid-info.com/ is a very good source of information for looking up the type of
    a specific OID

    :param rdnNameAttrValue: the value part of rfc3280.AttributeTypeAndValue

    :returns: the decoded value or raises PyAsn1Error if nothing worked
    """
    for decodeType in (
        asn1char.UTF8String,
        asn1char.PrintableString,
        asn1char.IA5String,
        asn1char.TeletexString,
        asn1char.BMPString,
    ):
        try:
            attrValStr, _rest = der_decode(rdnNameAttrValue, decodeType())
        # Decoding error, try the next type
        except PyAsn1Error:
            pass
        else:
            # If the decoding worked, return it
            return attrValStr
    raise PyAsn1Error("Could not find a correct decoding type")


def hasVOMSExtension(m2cert):
    """Utility fonction to check if the certificate has VOMS extensions

    :param m2cert: M2Crypto X509 object, a certificate

    :returns: boolean
    """
    try:
        retrieveExtension(m2cert, VOMS_EXTENSION_OID)
        return True
    except LookupError:
        return False


def decodeVOMSExtension(m2cert):
    """Decode the content of the VOMS extension

    :param m2cert: M2Crypto X509 object, a certificate

    :returns: A dictionnary containing the following fields:

      * notBefore: datetime.datetime
      * notAfter: datetime.datetime
      * attribute: (string). Comma separated list of VOMS tags presented as bellow

                             "<tagName> = <tagValue> (<tagQualifier>)"
                             Typically, the nickname will look like
                             'nickname = chaen (lhcb)',

      * fqan: List of VOMS "position" (['/lhcb/Role=production/Capability=NULL', '/lhcb/Role=NULL/Capability=NULL'])
      * vo: name of the VO,
      * subject: subject DN to which the attributes were granted,
      * issuer: typically the DN of the VOMS server (e.g '/DC=ch/DC=cern/OU=computers/CN=lcg-voms2.cern.ch')

    """
    vomsExtensionDict = {}
    vomsExtensionOctetString = retrieveExtension(m2cert, VOMS_EXTENSION_OID)
    # Decode it as a ACSequenceOfSequence, which is what it is...
    vomsExtensionSeqOfSeq, _rest = der_decode(vomsExtensionOctetString, asn1Spec=_ACSequenceOfSequence())

    # In principle, according to GFD 182, there could be more than one VO VOMS AC per proxy.
    # The standard specifies that we have to accept at least the first one, which is what
    # I will do...
    vomsCertAttribute = vomsExtensionSeqOfSeq[0][0]

    ######
    # TODO in principle, we should check the signature of the Attribute...
    # _signatureAlgorith = vomsCertAttribute['signatureAlgorithm']
    # _signatureValue = vomsCertAttribute['signatureValue']
    ######

    certAttrInfo = vomsCertAttribute["acinfo"]

    # pyasn1 does things correctly by setting a timezone info in the datetime
    # However, we do not in DIRAC, and so we can't compare the dates.
    # We have to remove the timezone info from the datetime objects

    notBefore = certAttrInfo["attrCertValidityPeriod"]["notBeforeTime"].asDateTime
    vomsExtensionDict["notBefore"] = notBefore.replace(tzinfo=None)

    notAfter = certAttrInfo["attrCertValidityPeriod"]["notAfterTime"].asDateTime
    vomsExtensionDict["notAfter"] = notAfter.replace(tzinfo=None)

    # ######### Retrieving the issuer ##########
    # Get the issuer. A bit tricky, because we have to reconstruct the full DN ourselves
    # The GFD 182 and RFC 3281 give enough restriction such that we can afford some direct
    # [0] access

    issuer = ""

    # rdnName is a rfc3280.RelativeDistinguishedName object
    for rdnName in certAttrInfo["issuer"]["v2Form"]["issuerName"][0]["directoryName"]["rdnSequence"]:
        # rdnNameAttr rfc3280.AttributeTypeAndValue'
        rdnNameAttr = rdnName[0]

        attrOid = ".".join([str(e) for e in rdnNameAttr["type"].asTuple()])

        # Now finally convert the last part into a asn1char.*String
        attrValStr = _decodeASN1String(rdnNameAttr["value"])
        attrVal = attrValStr.asOctets().decode()
        #
        issuer += "%s%s" % (DN_MAPPING[attrOid], attrVal)

    vomsExtensionDict["issuer"] = issuer

    # ### Issuer retrieved #####

    # ## Retrieving the Subject ####
    # We have to do the same for the subject than for the issuer

    subject = ""

    # rdnName is a rfc3280.RelativeDistinguishedName object
    for rdnName in certAttrInfo["holder"]["baseCertificateID"]["issuer"][0]["directoryName"]["rdnSequence"]:
        # rdnNameAttr rfc3280.AttributeTypeAndValue'
        rdnNameAttr = rdnName[0]

        attrOid = ".".join([str(e) for e in rdnNameAttr["type"].asTuple()])
        # # Because there are non printable characters in the values (new line, etc)
        # # we have to get ride of them. The best way is to get them as number, and make sure it is a
        # # a printable char (between 32 and 126)
        #
        # attrVal = ''.join([chr(c) for c in rdnNameAttr['value'].asNumbers() if 32 <= c <= 126 ])

        # Now finally convert the last part into a asn1char.*String
        attrValStr = _decodeASN1String(rdnNameAttr["value"])
        attrVal = attrValStr.asOctets().decode()

        subject += "%s%s" % (DN_MAPPING[attrOid], attrVal)

    vomsExtensionDict["subject"] = subject

    # ### Retrieving the FQAN ####

    # According to GFD182, there may be more attributes that just the FQAN, even though it
    # does not seem to be the case in practice. So we make sure to have the good one
    fqanOIDObj = univ.ObjectIdentifier(VOMS_FQANS_OID)

    # There shall be only one, hense the [0]
    # This is an rfc3280.Attribute object
    fqanAttrObj = [attrObj for attrObj in certAttrInfo["attributes"] if attrObj["type"] == fqanOIDObj][0]

    # According to GFD182 3.4.1, we decode the value as a IetfAttrSyntax.
    # Since multiple values are not allowed, just take the first item
    #
    fqanObj, _rest = der_decode(fqanAttrObj["values"][0], asn1Spec=rfc3281.IetfAttrSyntax())

    # We retrieve the VO and the VOMS server
    voName, _, _ = fqanObj["policyAuthority"][0]["uniformResourceIdentifier"].asOctets().decode().split(":")

    vomsExtensionDict["vo"] = voName

    # Now retrieve the position of the holder (group, role)
    fqanList = []
    for fqanPositionObj in fqanObj["values"]:
        fqanList.append(fqanPositionObj["octets"].asOctets().decode())

    vomsExtensionDict["fqan"] = fqanList

    # ############ End of the FQAN ################

    # Now the Tags, called attributes in the dict...

    tagDescriptions = []
    vomsTagsOIDObj = univ.ObjectIdentifier(VOMS_TAGS_EXT_OID)

    # First find the tag containers
    tagExtensionObj = [extObj for extObj in certAttrInfo["extensions"] if extObj["extnID"] == vomsTagsOIDObj]

    # If we found tags
    if tagExtensionObj:
        # Multiple is forbiden, so only one tag container
        tagExtensionObj = tagExtensionObj[0]

        tagContainersObj, _rest = der_decode(tagExtensionObj["extnValue"], asn1Spec=_TagContainers())

        # TODO in principle, we should check that this value
        # and the one of the policyAuthority of the fqan are the same
        # _tagPolicyAuthority = tagContainersObj[0][0]['policyAuthority'][0]['uniformResourceIdentifier'] \
        #     .asOctets().decode()
        ######

        for tagContainer in tagContainersObj:
            for tagList in tagContainer:
                # Note: it is here that I should check the policyAuthority
                tagList = tagList["tags"]
                for tag in tagList:
                    # This gives a string like
                    # nickname = chaen (lhcb)
                    tagDescriptions.append(
                        "%s = %s (%s)"
                        % (
                            tag["name"].asOctets().decode(),
                            tag["value"].asOctets().decode(),
                            tag["qualifier"].asOctets().decode(),
                        )
                    )

        vomsExtensionDict["attribute"] = ",".join(tagDescriptions)

    # #### Tags are done ################

    return vomsExtensionDict


def retrieveExtension(m2Cert, extensionOID):
    """Retrieves the extension from a certificate from its OID

    :param m2Cert: M2Crypto X509 object, a certificate
    :param extensionOID: the OID we are looking for

    :returns: an ~pyasn1.type.univ.OctetString object, which is the content of the extension
              (it still needs to be deserialized, depending on the extension !)

    :raises: LookupError if it does not have the extension
    """

    # Decode the certificate as a RFC2459 Certificate object.It is compatible
    # with the RFC proxy definition
    cert, _rest = der_decode(m2Cert.as_der(), asn1Spec=rfc2459.Certificate())
    extensions = cert["tbsCertificate"]["extensions"]

    # Construct an OID object for comparison purpose
    extensionOIDObj = univ.ObjectIdentifier(extensionOID)

    # We check every extension OID. This will be necessary until M2Crypto
    # allows to register OID alias (https://gitlab.com/m2crypto/m2crypto/issues/231)
    for extension in extensions:
        # We found the good extension
        if extension["extnID"] == extensionOIDObj:
            return extension["extnValue"]

    # If we are here, it means that we could not find the expected extension.
    raise LookupError("Could not find extension with OID %s" % extensionOID)



#!/usr/bin/env python
"""
Launch the File Catalog shell

Example:
  $ dirac-dms-filecatalog-cli
  Starting DIRAC FileCatalog client
  File Catalog Client $Revision: 1.17 $Date:
  FC:/>help

  Documented commands (type help <topic>):
  ========================================
  add    chmod  find   guid  ls     pwd       replicate  rmreplica   user
  cd     chown  get    id    meta   register  rm         size
  chgrp  exit   group  lcd   mkdir  replicas  rmdir      unregister

  Undocumented commands:
  ======================
  help

  FC:/>
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    fcType = "FileCatalog"
    Script.registerSwitch("f:", "file-catalog=", "   Catalog client type to use (default %s)" % fcType)
    Script.parseCommandLine(ignoreErrors=False)

    from DIRAC import gConfig, exit as dexit
    from DIRAC.Resources.Catalog.FileCatalogFactory import FileCatalogFactory

    __RCSID__ = "$Id$"

    fcType = gConfig.getValue("/LocalSite/FileCatalog", "")

    res = gConfig.getSections("/Resources/FileCatalogs", listOrdered=True)
    if not res["OK"]:
        dexit(1)
    fcList = res["Value"]
    if not fcType:
        if res["OK"]:
            fcType = res["Value"][0]

    for switch in Script.getUnprocessedSwitches():
        if switch[0].lower() == "f" or switch[0].lower() == "file-catalog":
            fcType = switch[1]

    if not fcType:
        print("No file catalog given and defaults could not be obtained")
        sys.exit(1)

    from DIRAC.DataManagementSystem.Client.FileCatalogClientCLI import FileCatalogClientCLI

    result = FileCatalogFactory().createCatalog(fcType)
    if not result["OK"]:
        print(result["Message"])
        if fcList:
            print("Possible choices are:")
            for fc in fcList:
                print(" " * 5, fc)
        sys.exit(1)
    print("Starting %s client" % fcType)
    catalog = result["Value"]
    cli = FileCatalogClientCLI(catalog)
    cli.cmdloop()


if __name__ == "__main__":
    main()



""" FileManager for ... ?
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import os
import datetime

from DIRAC import S_OK, S_ERROR
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.FileManager.FileManagerBase import FileManagerBase
from DIRAC.Core.Utilities.List import stringListToString, intListToString, breakListIntoChunks

# The logic of some methods is basically a copy/paste from the FileManager class,
# so I could have inherited from it. However, I did not want to depend on it


class FileManagerPs(FileManagerBase):
    def __init__(self, database=None):
        super(FileManagerPs, self).__init__(database)

    ######################################################
    #
    # The all important _findFiles and _getDirectoryFiles methods
    #

    def _findFiles(self, lfns, metadata=["FileID"], allStatus=False, connection=False):
        """Returns the information for the given lfns
        The logic works nicely in the FileManager, so I pretty much copied it.
        :param lfns: list of lfns
        :param metadata: list of params that we want to get for each lfn
        :param allStatus: consider all file status or only those defined in db.visibleFileStatus

        :return successful/failed convention. successful is a dict < lfn : dict of metadata >

        """
        connection = self._getConnection(connection)
        dirDict = self._getFileDirectories(lfns)

        result = self.db.dtree.findDirs(list(dirDict))
        if not result["OK"]:
            return result

        directoryIDs = result["Value"]

        failed = {}
        successful = {}
        for dirPath in directoryIDs:
            fileNames = dirDict[dirPath]
            res = self._getDirectoryFiles(
                directoryIDs[dirPath], fileNames, metadata, allStatus=allStatus, connection=connection
            )

            for fileName, fileDict in res.get("Value", {}).items():
                fname = os.path.join(dirPath, fileName)
                successful[fname] = fileDict

        # The lfns that are not in successful nor failed don't exist
        for failedLfn in set(lfns) - set(successful):
            failed.setdefault(failedLfn, "No such file or directory")

        return S_OK({"Successful": successful, "Failed": failed})

    def _findFileIDs(self, lfns, connection=False):
        """Find lfn <-> FileID correspondence"""
        connection = self._getConnection(connection)
        failed = {}
        successful = {}

        # If there is only one lfn, we might as well make a direct query
        if len(lfns) == 1:
            lfn = list(lfns)[0]  # if lfns is a dict, list(lfns) returns lfns.keys()
            pathPart, filePart = os.path.split(lfn)
            result = self.db.executeStoredProcedure(
                "ps_get_file_id_from_lfn", (pathPart, filePart, "ret1"), outputIds=[2]
            )
            if not result["OK"]:
                return result

            fileId = result["Value"][0]

            if not fileId:
                failed[lfn] = "No such file or directory"
            else:
                successful[lfn] = fileId

        else:

            # We separate the files by directory
            filesInDirDict = self._getFileDirectories(lfns)

            # We get the directory ids
            result = self.db.dtree.findDirs(list(filesInDirDict))
            if not result["OK"]:
                return result
            directoryPathToIds = result["Value"]

            # For each directory, we get the file ids of the files we want
            for dirPath in directoryPathToIds:
                fileNames = filesInDirDict[dirPath]
                dirID = directoryPathToIds[dirPath]

                formatedFileNames = stringListToString(fileNames)

                result = self.db.executeStoredProcedureWithCursor(
                    "ps_get_file_ids_from_dir_id", (dirID, formatedFileNames)
                )
                if not result["OK"]:
                    return result
                for fileID, fileName in result["Value"]:
                    fname = os.path.join(dirPath, fileName)
                    successful[fname] = fileID

            # The lfns that are not in successful dont exist
            for failedLfn in set(lfns) - set(successful):
                failed[failedLfn] = "No such file or directory"

        return S_OK({"Successful": successful, "Failed": failed})

    def _getDirectoryFiles(self, dirID, fileNames, metadata_input, allStatus=False, connection=False):
        """For a given directory, and eventually given file, returns all the desired metadata

        :param int dirID: directory ID
        :param fileNames: the list of filenames, or []
        :param metadata_input: list of desired metadata.
                   It can be anything from (FileName, DirID, FileID, Size, UID, Owner,
                   GID, OwnerGroup, Status, GUID, Checksum, ChecksumType, Type, CreationDate, ModificationDate, Mode)
        :param bool allStatus: if False, only displays the files whose status is in db.visibleFileStatus

        :returns: S_OK(files), where files is a dictionary indexed on filename, and values are dictionary of metadata
        """

        connection = self._getConnection(connection)

        metadata = list(metadata_input)
        if "UID" in metadata:
            metadata.append("Owner")
        if "GID" in metadata:
            metadata.append("OwnerGroup")
        if "FileID" not in metadata:
            metadata.append("FileID")

        # Format the filenames and status to be used in a IN clause in the sotred procedure
        formatedFileNames = stringListToString(fileNames)
        fStatus = stringListToString(self.db.visibleFileStatus)

        specificFiles = True if len(fileNames) else False
        result = self.db.executeStoredProcedureWithCursor(
            "ps_get_all_info_for_files_in_dir", (dirID, specificFiles, formatedFileNames, allStatus, fStatus)
        )

        if not result["OK"]:
            return result

        fieldNames = [
            "FileName",
            "DirID",
            "FileID",
            "Size",
            "UID",
            "Owner",
            "GID",
            "OwnerGroup",
            "Status",
            "GUID",
            "Checksum",
            "ChecksumType",
            "Type",
            "CreationDate",
            "ModificationDate",
            "Mode",
        ]

        rows = result["Value"]
        files = {}

        for row in rows:

            rowDict = dict(zip(fieldNames, row))
            fileName = rowDict["FileName"]
            # Returns only the required metadata
            files[fileName] = dict((key, rowDict.get(key, "Unknown metadata field")) for key in metadata)

        return S_OK(files)

    def _getFileMetadataByID(self, fileIDs, connection=False):
        """Get standard file metadata for a list of files specified by FileID

        :param fileIDS : list of file Ids

        :returns: S_OK(files), where files is a dictionary indexed on fileID
                            and the values dictionaries containing the following info:
                            ["FileID", "Size", "UID", "GID", "s.Status", "GUID", "CreationDate"]
        """

        # Format the filenames and status to be used in a IN clause in the sotred procedure
        formatedFileIds = intListToString(fileIDs)
        result = self.db.executeStoredProcedureWithCursor("ps_get_all_info_for_file_ids", (formatedFileIds,))
        if not result["OK"]:
            return result

        rows = result["Value"]

        fieldNames = ["FileID", "Size", "UID", "GID", "s.Status", "GUID", "CreationDate"]

        resultDict = {}

        for row in rows:
            rowDict = dict(zip(fieldNames, row))
            rowDict["Size"] = int(rowDict["Size"])
            rowDict["UID"] = int(rowDict["UID"])
            rowDict["GID"] = int(rowDict["GID"])
            resultDict[rowDict["FileID"]] = rowDict

        return S_OK(resultDict)

    def __insertMultipleFiles(self, allFileValues, wantedLfns):
        """Insert multiple files in one query. However, if there is a problem
            with one file, all the query is rolled back.
        :param allFileValues : dictionary of tuple with all the information about possibly more
                              files than we want to insert
        :param wantedLfns : list of lfn that we want to insert
        """

        fileValuesStrings = []
        fileDescStrings = []

        for lfn in wantedLfns:
            dirID, size, s_uid, s_gid, statusID, fileName, guid, checksum, checksumtype, mode = allFileValues[lfn]
            utcNow = datetime.datetime.utcnow().replace(microsecond=0)
            fileValuesStrings.append(
                "(%s, %s, %s, %s, %s, '%s', '%s', '%s', '%s', '%s', '%s', %s)"
                % (dirID, size, s_uid, s_gid, statusID, fileName, guid, checksum, checksumtype, utcNow, utcNow, mode)
            )
            fileDescStrings.append("(DirID = %s AND FileName = '%s')" % (dirID, fileName))

        fileValuesStr = ",".join(fileValuesStrings)
        fileDescStr = " OR ".join(fileDescStrings)

        result = self.db.executeStoredProcedureWithCursor("ps_insert_multiple_file", (fileValuesStr, fileDescStr))

        return result

    def __chunks(self, l, n):
        """Yield successive n-sized chunks from l."""
        for i in range(0, len(l), n):
            yield l[i : i + n]

    def _insertFiles(self, lfns, uid, gid, connection=False):
        """Insert new files. lfns is a dictionary indexed on lfn, the values are
        mandatory: DirID, Size, Checksum, GUID
        optional : Owner (dict with username and group), ChecksumType (Adler32 by default), Mode (db.umask by default)

        :param lfns : lfns and info to insert
        :param uid : user id, overwriten by Owner['username'] if defined
        :param gid : user id, overwriten by Owner['group'] if defined

        """

        connection = self._getConnection(connection)

        failed = {}
        successful = {}
        res = self._getStatusInt("AprioriGood", connection=connection)

        if res["OK"]:
            statusID = res["Value"]
        else:
            return res

        lfnsToRetry = []

        fileValues = {}
        fileDesc = {}

        # Prepare each file separately
        for lfn in lfns:

            # Get all the info
            fileInfo = lfns[lfn]

            dirID = fileInfo["DirID"]
            fileName = os.path.basename(lfn)
            size = fileInfo["Size"]
            ownerDict = fileInfo.get("Owner", None)
            checksum = fileInfo["Checksum"]
            checksumtype = fileInfo.get("ChecksumType", "Adler32")
            guid = fileInfo["GUID"]
            mode = fileInfo.get("Mode", self.db.umask)

            s_uid = uid
            s_gid = gid

            # overwrite the s_uid and s_gid if defined in the lfn info
            if ownerDict:
                result = self.db.ugManager.getUserAndGroupID(ownerDict)
                if result["OK"]:
                    s_uid, s_gid = result["Value"]

            fileValues[lfn] = (dirID, size, s_uid, s_gid, statusID, fileName, guid, checksum, checksumtype, mode)
            fileDesc[(dirID, fileName)] = lfn

        chunkSize = 200
        allChunks = list(self.__chunks(list(lfns), chunkSize))

        for lfnChunk in allChunks:
            result = self.__insertMultipleFiles(fileValues, lfnChunk)

            if result["OK"]:
                allIds = result["Value"]
                for dirId, fileName, fileID in allIds:
                    lfn = fileDesc[(dirId, fileName)]
                    successful[lfn] = lfns[lfn]
                    successful[lfn]["FileID"] = fileID
            else:
                lfnsToRetry.extend(lfnChunk)

        # If we are here, that means that the multiple insert failed, so we do one by one

        for lfn in lfnsToRetry:
            dirID, size, s_uid, s_gid, statusID, fileName, guid, checksum, checksumtype, mode = fileValues[lfn]
            # insert
            result = self.db.executeStoredProcedureWithCursor(
                "ps_insert_file", (dirID, size, s_uid, s_gid, statusID, fileName, guid, checksum, checksumtype, mode)
            )

            if not result["OK"]:
                failed[lfn] = result["Message"]
            else:

                fileID = result["Value"][0][0]

                successful[lfn] = lfns[lfn]
                successful[lfn]["FileID"] = fileID

        return S_OK({"Successful": successful, "Failed": failed})

    def _getFileIDFromGUID(self, guids, connection=False):
        """Returns the file ids from list of guids
        :param guids : list of guid

        :returns dictionary  < guid : fileId >

        """

        connection = self._getConnection(connection)
        if not guids:
            return S_OK({})

        if not isinstance(guids, (list, tuple)):
            guids = [guids]

        #     formatedGuids = ','.join( [ '"%s"' % guid for guid in guids ] )
        formatedGuids = stringListToString(guids)
        result = self.db.executeStoredProcedureWithCursor("ps_get_file_ids_from_guids", (formatedGuids,))

        if not result["OK"]:
            return result

        guidDict = dict((guid, fileID) for guid, fileID in result["Value"])

        return S_OK(guidDict)

    def getLFNForGUID(self, guids, connection=False):
        """Returns the lfns matching given guids"""
        connection = self._getConnection(connection)
        if not guids:
            return S_OK({})

        if not isinstance(guids, (list, tuple)):
            guids = [guids]

        formatedGuids = stringListToString(guids)
        result = self.db.executeStoredProcedureWithCursor("ps_get_lfns_from_guids", (formatedGuids,))

        if not result["OK"]:
            return result

        guidDict = dict((guid, lfn) for guid, lfn in result["Value"])
        failedGuid = set(guids) - set(guidDict)
        failed = dict.fromkeys(failedGuid, "GUID does not exist") if failedGuid else {}
        return S_OK({"Successful": guidDict, "Failed": failed})

    ######################################################
    #
    # _deleteFiles related methods
    #

    def _deleteFiles(self, fileIDs, connection=False):
        """Delete a list of files and the associated replicas

        :param fileIDS : list of fileID

        :returns: S_OK() or S_ERROR(msg)
        """

        connection = self._getConnection(connection)

        replicaPurge = self.__deleteFileReplicas(fileIDs)
        filePurge = self.__deleteFiles(fileIDs, connection=connection)

        if not replicaPurge["OK"]:
            return replicaPurge

        if not filePurge["OK"]:
            return filePurge

        return S_OK()

    def __deleteFileReplicas(self, fileIDs, connection=False):
        """Delete all the replicas from the file ids

        :param fileIDs: list of file ids

        :returns: S_OK() or S_ERROR(msg)
        """

        connection = self._getConnection(connection)

        if not fileIDs:
            return S_OK()

        formatedFileIds = intListToString(fileIDs)

        result = self.db.executeStoredProcedureWithCursor("ps_delete_replicas_from_file_ids", (formatedFileIds,))
        if not result["OK"]:
            return result

        errno, msg = result["Value"][0]

        if errno:
            return S_ERROR(msg)

        return S_OK()

    def __deleteFiles(self, fileIDs, connection=False):
        """Delete the files from their ids

        :param fileIDs: list of file ids

        :returns: S_OK() or S_ERROR(msg)
        """

        connection = self._getConnection(connection)

        formatedFileIds = intListToString(fileIDs)

        result = self.db.executeStoredProcedureWithCursor("ps_delete_files", (formatedFileIds,))
        if not result["OK"]:
            return result

        errno, msg = result["Value"][0]

        if errno:
            return S_ERROR(msg)

        return S_OK()

    def __insertMultipleReplicas(self, allReplicaValues, lfnsChunk):
        """Insert multiple replicas in one query. However, if there is a problem
            with one replica, all the query is rolled back.
        :param allReplicaValues : dictionary of tuple with all the information about possibly more
                              replica than we want to insert
        :param lfnsChunk : list of lfn that we want to insert
        """

        repValuesStrings = []
        repDescStrings = []

        for lfn in lfnsChunk:
            fileID, seID, statusID, replicaType, pfn = allReplicaValues[lfn]
            utcNow = datetime.datetime.utcnow().replace(microsecond=0)
            repValuesStrings.append(
                "(%s,%s,'%s','%s','%s','%s','%s')" % (fileID, seID, statusID, replicaType, utcNow, utcNow, pfn)
            )
            repDescStrings.append("(r.FileID = %s AND SEID = %s)" % (fileID, seID))

        repValuesStr = ",".join(repValuesStrings)
        repDescStr = " OR ".join(repDescStrings)

        result = self.db.executeStoredProcedureWithCursor("ps_insert_multiple_replica", (repValuesStr, repDescStr))

        return result

    def _insertReplicas(self, lfns, master=False, connection=False):
        """Insert new replicas. lfns is a dictionary with one entry for each file. The keys are lfns, and values are dict
        with mandatory attributes : FileID, SE (the name), PFN

        :param lfns: lfns and info to insert
        :param master: true if they are master replica, otherwise they will be just 'Replica'

        :return: successful/failed convention, with successful[lfn] = true
        """
        chunkSize = 200

        connection = self._getConnection(connection)

        # Add the files
        failed = {}
        successful = {}

        # Get the status id of AprioriGood
        res = self._getStatusInt("AprioriGood", connection=connection)
        if not res["OK"]:
            return res
        statusID = res["Value"]

        lfnsToRetry = []

        repValues = {}
        repDesc = {}

        # treat each file after each other
        for lfn in lfns.keys():

            fileID = lfns[lfn]["FileID"]

            seName = lfns[lfn]["SE"]
            if isinstance(seName, six.string_types):
                seList = [seName]
            elif isinstance(seName, list):
                seList = seName
            else:
                return S_ERROR("Illegal type of SE list: %s" % str(type(seName)))

            replicaType = "Master" if master else "Replica"
            pfn = lfns[lfn]["PFN"]

            # treat each replica of a file after the other
            # (THIS CANNOT WORK... WE ARE ONLY CAPABLE OF DOING ONE REPLICA PER FILE AT THE TIME)
            for seName in seList:

                # get the SE id
                res = self.db.seManager.findSE(seName)
                if not res["OK"]:
                    failed[lfn] = res["Message"]
                    continue
                seID = res["Value"]

                # This is incompatible with adding multiple replica at the time for a given file
                repValues[lfn] = (fileID, seID, statusID, replicaType, pfn)
                repDesc[(fileID, seID)] = lfn

        allChunks = list(self.__chunks(list(lfns), chunkSize))

        for lfnChunk in allChunks:
            result = self.__insertMultipleReplicas(repValues, lfnChunk)

            if result["OK"]:
                allIds = result["Value"]
                for fileId, seId, repId in allIds:
                    lfn = repDesc[(fileId, seId)]
                    successful[lfn] = True
                    lfns[lfn]["RepID"] = repId
            else:
                lfnsToRetry.extend(lfnChunk)

        for lfn in lfnsToRetry:
            fileID, seID, statusID, replicaType, pfn = repValues[lfn]
            # insert the replica and its info
            result = self.db.executeStoredProcedureWithCursor(
                "ps_insert_replica", (fileID, seID, statusID, replicaType, pfn)
            )

            if not result["OK"]:
                failed[lfn] = result["Message"]
            else:
                replicaID = result["Value"][0][0]
                lfns[lfn]["RepID"] = replicaID
                successful[lfn] = True

        return S_OK({"Successful": successful, "Failed": failed})

    def _getRepIDsForReplica(self, replicaTuples, connection=False):
        """Get the Replica IDs for (fileId, SEID) couples

        :param repliacTuples : list of (fileId, SEID) couple

        :returns { fileID : { seID : RepID } }
        """
        connection = self._getConnection(connection)

        replicaDict = {}

        for fileID, seID in replicaTuples:
            result = self.db.executeStoredProcedure("ps_get_replica_id", (fileID, seID, "repIdOut"), outputIds=[2])
            if not result["OK"]:
                return result

            repID = result["Value"][0]

            # if the replica exists, we add it to the dict
            if repID:
                replicaDict.setdefault(fileID, {}).setdefault(seID, repID)

        return S_OK(replicaDict)

    ######################################################
    #
    # _deleteReplicas related methods
    #

    def _deleteReplicas(self, lfns, connection=False):
        """Deletes replicas. The deletion of replicas that do not exist is successful

        :param lfns : dictinary with lfns as key, and the value is a dict with a mandatory "SE" key,
                      corresponding to the SE name or SE ID

        :returns: successful/failed convention, with successful[lfn] = True
        """
        connection = self._getConnection(connection)
        failed = {}
        successful = {}
        # First we get the fileIds from our lfns
        res = self._findFiles(list(lfns), ["FileID"], connection=connection)
        if not res["OK"]:
            return res

        # If the file does not exist we consider the deletion successful
        for lfn, error in res["Value"]["Failed"].items():
            if error == "No such file or directory":
                successful[lfn] = True
            else:
                failed[lfn] = error

        lfnFileIDDict = res["Value"]["Successful"]
        for lfn, fileDict in lfnFileIDDict.items():
            fileID = fileDict["FileID"]

            # Then we get our StorageElement Id (cached in seManager)
            se = lfns[lfn]["SE"]
            # if se is already the se id, findSE will return it
            res = self.db.seManager.findSE(se)
            if not res["OK"]:
                return res
            seID = res["Value"]

            # Finally remove the replica
            result = self.db.executeStoredProcedureWithCursor("ps_delete_replica_from_file_and_se_ids", (fileID, seID))
            if not result["OK"]:
                failed[lfn] = result["Message"]
                continue

            errno, errMsg = result["Value"][0]
            if errno:
                failed[lfn] = errMsg
            else:
                successful[lfn] = True

        return S_OK({"Successful": successful, "Failed": failed})

    ######################################################
    #
    # _setReplicaStatus _setReplicaHost _setReplicaParameter methods
    # _setFileParameter method
    #

    def _setReplicaStatus(self, fileID, se, status, connection=False):
        """Set the status of a replica

        :param fileID : file id
        :param se : se name or se id
        :param status : status to be applied

        :returns: S_OK() or S_ERROR(msg)
        """
        if status not in self.db.validReplicaStatus:
            return S_ERROR("Invalid replica status %s" % status)
        connection = self._getConnection(connection)
        res = self._getStatusInt(status, connection=connection)
        if not res["OK"]:
            return res
        statusID = res["Value"]

        # Then we get our StorageElement Id (cached in seManager)
        res = self.db.seManager.findSE(se)
        if not res["OK"]:
            return res
        seID = res["Value"]

        result = self.db.executeStoredProcedureWithCursor("ps_set_replica_status", (fileID, seID, statusID))
        if not result["OK"]:
            return result

        affected = result["Value"][0][0]  # Affected is the number of raws updated

        if not affected:
            return S_ERROR("Replica does not exist")
        return S_OK()

    def _setReplicaHost(self, fileID, se, newSE, connection=False):
        """Move a replica from one SE to another (I don't think this should be called

        :param fileID : file id
        :param se : se name or se id of the previous se
        :param newSE : se name or se id of the new se

        :returns: S_OK() or S_ERROR(msg)
        """
        connection = self._getConnection(connection)

        # Get the new se id
        res = self.db.seManager.findSE(newSE)
        if not res["OK"]:
            return res
        newSEID = res["Value"]

        # Get the old se id
        res = self.db.seManager.findSE(se)
        if not res["OK"]:
            return res
        oldSEID = res["Value"]

        # update
        result = self.db.executeStoredProcedureWithCursor("ps_set_replica_host", (fileID, oldSEID, newSEID))
        if not result["OK"]:
            return result

        affected = result["Value"][0][0]
        if not affected:
            return S_ERROR("Replica does not exist")
        else:
            return S_OK()

    def _setFileParameter(self, fileID, paramName, paramValue, connection=False):
        """Generic method to set a file parameter


        :param fileID : id of the file
        :param paramName : the file parameter you want to change
              It should be one of [ UID, GID, Status, Mode]. However, in case of
              unexpected parameter, and to stay compatible with the other Manager,
              there is a manual request done.
        :param paramValue : the value (raw, or id) to insert

        :returns: S_OK() or S_ERROR

        """
        connection = self._getConnection(connection)

        # The PS associated with a given parameter
        psNames = {
            "UID": "ps_set_file_uid",
            "GID": "ps_set_file_gid",
            "Status": "ps_set_file_status",
            "Mode": "ps_set_file_mode",
        }

        psName = psNames.get(paramName, None)

        # If there is an associated procedure, we go for it
        if psName:

            result = self.db.executeStoredProcedureWithCursor(psName, (fileID, paramValue))
            if not result["OK"]:
                return result

            _affected = result["Value"][0][0]
            # If affected = 0, the file does not exist, but who cares...

        # In case this is a 'new' parameter, we have a failback solution, but we
        # should add a specific ps for it
        else:
            req = "UPDATE FC_Files SET %s='%s', ModificationDate=UTC_TIMESTAMP() WHERE FileID IN (%s)" % (
                paramName,
                paramValue,
                intListToString(fileID),
            )
            return self.db._update(req, connection)

        return S_OK()

    ######################################################
    #
    # _getFileReplicas related methods
    #

    def _getFileReplicas(self, fileIDs, fields_input=None, allStatus=False, connection=False):
        """Get replicas for the given list of files specified by their fileIDs
        :param fileIDs : list of file ids
        :param fields_input : metadata of the Replicas we are interested in (default to PFN)
        :param allStatus : if True, all the Replica statuses will be considered,
                           otherwise, only the db.visibleReplicaStatus

        :returns S_OK with a dict { fileID : { SE name : dict of metadata } }
        """

        if fields_input is None:
            fields_input = ["PFN"]

        fields = list(fields_input)

        # always add Status in the list of required fields
        if "Status" not in fields:
            fields.append("Status")

        # We initialize the dictionary with empty dict
        # as default value, because this is what we want for
        # non existing replicas
        replicas = {fileID: {} for fileID in fileIDs}

        # Format the status to be used in a IN clause in the stored procedure
        fStatus = stringListToString(self.db.visibleReplicaStatus)

        fieldNames = ["FileID", "SE", "Status", "RepType", "CreationDate", "ModificationDate", "PFN"]

        for chunks in breakListIntoChunks(fileIDs, 1000):
            # Format the FileIDs to be used in a IN clause in the stored procedure
            formatedFileIds = intListToString(chunks)
            result = self.db.executeStoredProcedureWithCursor(
                "ps_get_all_info_of_replicas_bulk", (formatedFileIds, allStatus, fStatus)
            )

            if not result["OK"]:
                return result

            rows = result["Value"]

            for row in rows:
                rowDict = dict(zip(fieldNames, row))
                se = rowDict["SE"]
                fileID = rowDict["FileID"]
                replicas[fileID][se] = dict((key, rowDict.get(key, "Unknown metadata field")) for key in fields)

        return S_OK(replicas)

    def countFilesInDir(self, dirId):
        """Count how many files there is in a given Directory

        :param dirID: directory id

        :returns: S_OK(value) or S_ERROR
        """

        result = self.db.executeStoredProcedure("ps_count_files_in_dir", (dirId, "ret1"), outputIds=[1])
        if not result["OK"]:
            return result

        res = S_OK(result["Value"][0])
        return res

    ##########################################################################
    #
    #  We overwrite some methods from the base class because of the new DB constraints or perf reasons
    #
    #  Some methods could be inherited in the future if we have perf problems. For example
    #  * setFileGroup
    #  * setFileOwner
    #  * setFileMode
    #  * changePath*
    #
    ##########################################################################

    def _updateDirectoryUsage(self, directorySEDict, change, connection=False):
        """This updates the directory usage, but is now done by triggers in the DB"""
        return S_OK()

    def _computeStorageUsageOnRemoveFile(self, lfns, connection=False):
        """Again nothing to compute, all done by the triggers"""
        directorySESizeDict = {}
        return S_OK(directorySESizeDict)

    #   "REMARQUE : THIS IS STILL TRUE, BUT YOU MIGHT WANT TO CHECK FOR A GIVEN GUID ANYWAY
    #   def _checkUniqueGUID( self, lfns, connection = False ):
    #     """ The GUID unicity is ensured at the DB level, so we will have similar message if the insertion fails"""
    #
    #     failed = {}
    #     return failed

    def getDirectoryReplicas(self, dirID, path, allStatus=False, connection=False):
        """
        This is defined in the FileManagerBase but it relies on the SEManager to get the SE names.
        It is good practice in software, but since the SE and Replica tables are bound together in the DB,
        I might as well resolve the name in the query


        Get the replicas for all the Files in the given Directory

        :param int dirID: ID of the directory
        :param unused path: useless
        :param bool allStatus: whether all replicas and file status are considered
                               If False, take the visibleFileStatus and visibleReplicaStatus
                               values from the configuration
        """

        # We format the visible file/replica satus so we can give it as argument to the ps
        # It is used in an IN clause, so it looks like --'"AprioriGood","Trash"'--
        #     fStatus = ','.join( [ '"%s"' % status for status in self.db.visibleFileStatus ] )
        #     rStatus = ','.join( [ '"%s"' % status for status in self.db.visibleReplicaStatus ] )
        fStatus = stringListToString(self.db.visibleFileStatus)
        rStatus = stringListToString(self.db.visibleReplicaStatus)

        result = self.db.executeStoredProcedureWithCursor(
            "ps_get_replicas_for_files_in_dir", (dirID, allStatus, fStatus, rStatus)
        )
        if not result["OK"]:
            return result

        resultDict = {}
        for fileName, _fileID, seName, pfn in result["Value"]:
            resultDict.setdefault(fileName, {}).setdefault(seName, []).append(pfn)

        return S_OK(resultDict)

    def _getFileLFNs(self, fileIDs):
        """Get the file LFNs for a given list of file IDs
        We need to override this method because the base class hard codes the column names
        """

        successful = {}
        for chunks in breakListIntoChunks(fileIDs, 1000):
            # Format the filenames and status to be used in a IN clause in the sotred procedure
            formatedFileIds = intListToString(chunks)
            result = self.db.executeStoredProcedureWithCursor("ps_get_full_lfn_for_file_ids", (formatedFileIds,))
            if not result["OK"]:
                return result

            # The result contains FileID, LFN
            for row in result["Value"]:
                successful[row[0]] = row[1]

        missingIds = set(fileIDs) - set(successful)
        failed = dict.fromkeys(missingIds, "File ID not found")

        return S_OK({"Successful": successful, "Failed": failed})

    def getSEDump(self, seName):
        """
         Return all the files at a given SE, together with checksum and size

        :param seName: name of the StorageElement

        :returns: S_OK with list of tuples (lfn, checksum, size)
        """

        res = self.db.seManager.findSE(seName)
        if not res["OK"]:
            return res
        seID = res["Value"]

        return self.db.executeStoredProcedureWithCursor("ps_get_se_dump", (seID,))
